# -*- coding: utf-8 -*-
"""
Created on Mon Jan 29 21:43:26 2018

@author: Callie Darsey

Canny Edge Detection and Hough Line Tranform example
"""

import cv2
import numpy as np

def calc_x(y, rho0, theta0):
    '''
    Coordinate System of four quadrants
    
    III | IV
    --------- x
    II  |  I
        y
    
    Given the y value, return the x value that places us on the line
    defined by theta0 and rho0.
 
    0 <= theta0 < 180 degrees, measured clock wise from the positive x axis.
    
    The value rho0, which can be positive or negative, gives the direction
    and distance at which the ray, defined by theta0 and the origin, has an
    perpendicular intersection with the line. 
    '''

    c = np.cos(theta0)
    s = np.sin(theta0)
    
    x0 = rho0 * c
    y0 = rho0 * s

    x = - s / c * y + (x0 + s / c * y0)
    
    return x

def calc_y(x, rho0, theta0):
    '''
    Coordinate System of four quadrants
    
    III | IV
    --------- x
    II  |  I
        y
    
    Given the x value, return the y value that places us on the line
    defined by theta0 and rho0.
 
    0 <= theta0 < 180 degrees, measured clock wise from the positive x axis.
    
    The value rho0, which can be positive or negative, gives the direction
    and distance at which the ray, defined by theta0 and the origin, has an
    perpendicular intersection with the line. 
    '''    
    c = np.cos(theta0)
    s = np.sin(theta0)
    
    x0 = rho0 * c
    y0 = rho0 * s

    y = - c / s * x + (y0 + c / s * x0)

    return y

def draw_hough_lines_like_example(img, theta, rho):
    '''
    This function returns the img with lines drawn in it according 
    to this example:
        
    https://docs.opencv.org/3.0-beta/doc/py_tutorials/py_imgproc/py_houghlines/py_houghlines.html

    img - image we want to drqw the lines in
    theta - theta defining a line with rho
    rho - rho defining a line with theta
    verbose - if this is True, then print statements are made    
    '''
    a = np.cos(theta)
    b = np.sin(theta)
    x0 = a*rho
    y0 = b*rho

    x1 = int(x0 + 1000*(-b))
    y1 = int(y0 + 1000*(a))
    x2 = int(x0 - 1000*(-b))
    y2 = int(y0 - 1000*(a))

    cv2.line(img_copy,(x1,y1),(x2,y2), (0, 0, 255), 1)
    
    return img



def draw_hough_lines(img, theta, rho, i, verbose):
    '''
    This function returns the img with lines drawn in it.
    
    img - image we want to drqw the lines in
    theta - theta defining a line with rho
    rho - rho defining a line with theta
    i - index to the line
    verbose - if this is True, then print statements are made
    '''

    c = np.cos(theta)
    s = np.sin(theta)
    
    if rho * c > 0.0 and rho * s > 0.0:
        
        quadrant = 'I'        
        
        x1 = 0
        y1 = calc_y(x1, rho, theta)
        
        if y1 > img.shape[0] - 1:
            
            y1 = img.shape[0] - 1
            x1 = calc_x(y1, rho, theta)
        
        x2 = img.shape[1] - 1
        y2 = calc_y(x2, rho, theta)
        
        if y2 < 0:
            
            y2 = 0
            x2 = calc_x(y2, rho, theta)
            
    elif rho * c < 0.0 and rho * s > 0.0:
        
        quadrant = 'II'
        
        x1 = 0
        y1 = calc_y(x1, rho, theta)
        
        x2 = img.shape[1] - 1
        y2 = calc_y(x2, rho, theta)
        
        if y2 > img.shape[0] - 1:
            
            y2 = img.shape[0] - 1
            x2 = calc_x(y2, rho, theta)
     
    elif rho * c < 0.0 and rho * s < 0.0:
        
        quadrant = 'III'
        print('Figured quadrant', quadrant, "would never happen. What's this?")
        
    elif rho * c > 0.0 and rho * s < 0.0:
        
        quadrant = 'IV'
        
        y1 = 0
        x1 = calc_x(y1, rho, theta)
        
        x2 = img.shape[1] - 1
        y2 = calc_y(x2, rho, theta)
        
        if y2 > img.shape[0] - 1:
            
            y2 = img.shape[0] - 1
            x2 = calc_x(y2, rho, theta)

    elif s == 0.0:
        
        quadrant = 'horizontal line'
        
        x1 = rho
        y1 = 0
        
        x2 = rho
        y2 = img.shape[0] - 1
        
    elif c == 0.0:
        
        quadrant = 'vertical line'
        
        x1 = 0
        y1 = rho
        
        x2 = img.shape[1] - 1
        y2 = rho
        
    else:
        
        quadrant = ''

        print('\nDid not fit in a category\n')
        
        x1 = -1
        y1 = 0
        
        x2 = 0
        y2 = -1
        
    x1_int = int(x1)
    y1_int = int(y1)
    x2_int = int(x2)
    y2_int = int(y2)
    
    if verbose == True:
        print('i = ', i, 'quadrant', quadrant, \
              (x1, y1), (x2, y2), (x1_int, y1_int), (x2_int, y2_int))

    cv2.line(img,(x1_int,y1_int),(x2_int,y2_int),colors[i],1)
    
    return img
    
    
img = cv2.imread('sudoku-original.jpg')
img_copy = np.copy(img)
img_copy2 = np.copy(img)
gray = cv2.cvtColor(img,cv2.COLOR_BGR2GRAY)
edges = cv2.Canny(gray,50,150,apertureSize = 3)

cv2.imwrite('hough_input_canny.jpg',edges)

lines = cv2.HoughLines(edges,1,np.pi/180,200)

colors = [  (255, 0, 0),      (0, 255, 0),     (0, 0, 255),   (255, 255, 0),
          (255, 0, 255),    (0, 255, 255),     (127, 0, 0),     (0, 127, 0),
            (0, 0, 127),    (127, 127, 0),   (127, 0, 127),   (0, 127, 127),
        (100, 100, 200),  (100, 200, 100), (200, 100, 100),    (127, 50, 0),
           (50, 127, 0),     (127, 0, 50),    (0, 50, 127), (255, 255, 100)]

rlines = np.reshape(lines, (20, 2))





for i, parameters in enumerate(rlines):
    
    rho, theta = parameters
    print('(theta, rho)', (theta, rho))
    
    img_copy = draw_hough_lines(img_copy, theta, rho, i, verbose=True)
    
    draw_hough_lines_like_example(img, theta, rho)

cv2.imwrite('houghlines3.jpg',img)
cv2.imwrite('houghlines3_copy.jpg', img_copy)


minLineLength = 100
maxLineGap = 10
lines = cv2.HoughLinesP(edges,1,np.pi/180,100,minLineLength,maxLineGap)
for x1,y1,x2,y2 in np.reshape(lines, (lines.shape[0], lines.shape[2])):
    cv2.line(img_copy2,(x1,y1),(x2,y2),(0,255,0), 1)

cv2.imwrite('houghlines5.jpg',img_copy2)


