# -*- coding: utf-8 -*-
"""
Created on Mon Apr 23 15:11:26 2018

@author: Callie Darsey


Center and rotate a blob

"""

import cv2
import numpy as np
import pickle
import matplotlib.pyplot as plt
import math

def fill_holes(mask):
    '''
    Fill the holes in connected components in the mask.
    
    References:
        https://www.learnopencv.com/filling-holes-in-an-image-using-opencv-python-c/
        
        http://answers.opencv.org/question/9863/fill-holes-of-a-binary-image/
    '''

    assert mask.dtype == np.uint8, 'fill_holes: mask expected to be UINT8'
    
    # force nonzero values to be 255
    mask[mask != 0] = 255
    
    # make a zero border around image    
    mask_floodfill = np.zeros((mask.shape[0] + 2, mask.shape[1] + 2),
                              dtype=np.uint8)
    mask_floodfill[1:mask.shape[0] + 1, 1:mask.shape[1] + 1] = mask
    
    # starting from the zero border, do a flood fill
    cv2.floodFill(mask_floodfill, None, (0, 0), 255)
    
    # invert making the holes in the connected components 255
    holes = cv2.bitwise_not(mask_floodfill[1:mask.shape[0] + 1,
                                           1:mask.shape[1] + 1])
    
    # fill the holes via OR and return
    return cv2.bitwise_or(mask, holes)


def centroid_center_crop(frame, cc, cc_label):
    '''
    This function crops connected component(s) such that the centroid is in 
    the middle of the crop (zero padding is added to achieve this if
    necessary).
    
    frame - frame the mask will extract information from
    cc - list holding output of cv2.connectedComponentsWithStats with
         any appended information 
    cc_label - connected component label being cropped
    '''
    
    # get out the one component
    one_cc = np.zeros_like(cc[1])
    one_cc[cc[1] == cc_label] = cc_label
    
    # mask the portion of frame aligned with the connected component
    portion = cv2.bitwise_and(frame, frame, mask=(one_cc > 0).astype(np.uint8))
    
    # centroid x and y forced to int
    c_col = int(cc[3][cc_label, 0])
    c_row = int(cc[3][cc_label, 1])
    
    # points of the bounding box.
    pt1_col = cc[2][cc_label, cv2.CC_STAT_LEFT]
    pt1_row = cc[2][cc_label, cv2.CC_STAT_TOP]
    
    pt2_col = pt1_col + cc[2][cc_label, cv2.CC_STAT_WIDTH] - 1
    pt2_row = pt1_row + cc[2][cc_label, cv2.CC_STAT_HEIGHT] - 1
    
    # left and right column distance
    left_col_dist = c_col - pt1_col
    right_col_dist = pt2_col - c_col
    
    # left and right row distance
    upper_row_dist = c_row - pt1_row
    lower_row_dist = pt2_row - c_row
    
    # row and column with of one side of rectangle excluding center
    col_dist = max([left_col_dist, right_col_dist])
    row_dist = max([upper_row_dist, lower_row_dist])
    
    # extract the portion of the frame associated with connected component
    extracted = portion[pt1_row:pt1_row + cc[2][cc_label, cv2.CC_STAT_HEIGHT],
                        pt1_col:pt1_col + cc[2][cc_label, cv2.CC_STAT_WIDTH]]
    
    # place the croped component with its centroid in the center
    crop = np.zeros((2 * row_dist + 1, 2 * col_dist + 1), dtype=np.uint8)
    crop[row_dist - upper_row_dist : \
         row_dist - upper_row_dist + cc[2][cc_label, cv2.CC_STAT_HEIGHT], \
         col_dist - left_col_dist : \
         col_dist - left_col_dist + cc[2][cc_label, cv2.CC_STAT_WIDTH], ...] = \
         extracted
         
    return crop
         
    
    
    
def insert_centroid_centered_crop(centroid_centered_crop, new_dim):
    '''
    Insert image into a new image of dimension new_dim.
    
    centroid_centered_crop - crop image with the centroid at the center
    new_dim - 2 element or three element size for the new image (the
              dimensions). First two dimensions are expected to be odd.
    '''
    
    assert len(centroid_centered_crop.shape) == len(new_dim)
    assert new_dim[0] % 2 == 1 and new_dim[1] % 2 == 1
    assert centroid_centered_crop.shape[0] % 2 == 1
    assert centroid_centered_crop.shape[1] % 2 == 1
    
    old_height, old_width = centroid_centered_crop.shape[0:2]
    
    new_height, new_width = new_dim[0:2]
    
    new = np.zeros(new_dim, dtype=np.uint8)
    
    new[int(new_height/2) - int(old_height/2) : \
        int(new_height/2) - int(old_height/2) + old_height,
        int(new_width/2) - int(old_width/2) : \
        int(new_width/2) - int(old_width/2) + old_width, ...] = \
        centroid_centered_crop
        
    return new

with open('./hand.data', 'rb') as f:
    obj_data, frame = pickle.load(f)
    
frame_gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)

mask = (obj_data['hand0001764995'] * frame_gray).astype(np.uint8) * 255

plt.imshow(mask)
plt.show()

mask = fill_holes(mask)

plt.imshow(mask)
plt.show()

kernel=np.ones((3,3))
mask = cv2.morphologyEx(mask, cv2.MORPH_OPEN, kernel, iterations=2)

plt.imshow(mask)
plt.show()

rows, cols = np.where(mask > 0)

# reshape the points into N x 2 for passing to cv2.minAreaRect()
points = np.stack([cols, rows], axis=1)

# get the minimum area rectangle info
rect_center_point, rect_shape, rotation = cv2.minAreaRect(points)

# separate the rotations
width_along_rotated_x, height_along_rotated_y = rect_shape

print('width_along_rotated_x', width_along_rotated_x)
print('height_along_rotated_y', height_along_rotated_y)
print('rectangle shape', rect_shape)
print('rotation', rotation)


'''when width_along_rotated_x is smaller than height_along_rotated_y, 
adjust the rotation to distinguish orientation'''
if width_along_rotated_x < height_along_rotated_y:
    
    if abs(rotation) != 90:
        
        rotation = -rotation
        
    else:
        
        rotation = 0
        
#-force -90 to become 90'''            
elif rotation == -90:
    
    rotation = 90

# save orientation
print(rotation)

'''
           negative number of degrees from x axis
         /
        /
       /
       ---------------> x
       |\
       | \
       |  \ positive number of degrees from x axis
       |
       y
'''


'''
https://stackoverflow.com/questions/35854197/how-to-use-opencvs-connected-components-with-stats-in-python
'''
cc = cv2.connectedComponentsWithStats(mask)

centered_crop = centroid_center_crop(mask, cc, 1)

plt.imshow(centered_crop)
plt.show()

biggest_dim = np.max(centered_crop.shape[0:2])

padded = \
    insert_centroid_centered_crop(centered_crop, (2 * biggest_dim + 1,
                                                  2 * biggest_dim + 1))
plt.imshow(padded)
plt.show()

center = int((2 * biggest_dim + 1) / 2)
    
M = cv2.getRotationMatrix2D((center, center), rotation, 1)
         
padded_rot = cv2.warpAffine(padded, M, padded.shape)

plt.imshow(padded_rot)
plt.show()

# remove padding
cc_padded_rot = cv2.connectedComponentsWithStats(padded_rot)
centered_rotated = centroid_center_crop(padded_rot, cc_padded_rot, 1)

plt.imshow(centered_rotated)
plt.show()



