# -*- coding: utf-8 -*-
"""
Created on Sat Feb 17 15:57:18 2018

@author: Callie Darsey

Given  the resized color frame, its flow, and exemplar(s) information for the
frame, form the three versions of  keep (1) / toss (0) components

component_exploration_02
"""



import pickle
import numpy as np
import cv2
import skin_detector_01 as skin_detector
from matplotlib import pyplot as plt
import fns_investigating_optical_flow_01 as fns
import overlay_01 as overlay
import component_functions_01 as comp_fns
import example_of_DFT_01 as dft_hist_fns

def get_cc_info(frame, flow, boxes):
    '''
    
    This function returns three tuples for the three versions of keep (1) and
    toss (0) components. It also returns the thresholded motion mask.
    
    The three ways are as follows:
        
        * WAY #1 for components found via
              1) finding the intersection (AND) between the skin detection and
                 thresholded motion
              2) taking the distance transform (DT or dt) on the intersection
                 and thresholding the DT at some chosen threshold value
              3) finding the intersection (AND) of the DT thresholded and the
                 thresholded motion
                 
        * WAY #2 for components found via        
              1) find the components according to WAY #1
              2) find watershed lines using the distance transform on WAY #1
                 components to mark the watershed function marker input
              2) break the WAY #1 components with the dilated watershed
                 (WS or ws) marker lines
                 
        * WAY #3 for components found via
              1) use the dilated watershed markers lines of WAY #2 and reject
                 the largest watershed (background) component
              2) if a watershed component overlaps a component of WAY #1 at
                 all, use as component
    
    Each of the returned components has the cv2.connectedComponentsWithStats()
    output with an appended element holding the label 1 for 'keep' and 0 for
    'toss' and -1 for the background component with label 0. Furthermore, it
    has lists of the keep labels and toss labels.
    
    frame - BGR color image
    flow - two channel flow image
    boxes - number of exemplars by four values for point 1 and point 2
            defining exemplar box
    '''
    

        
    '''WAY #1 for components (no watershed)'''
    
    # make exemplars mask
    exemplars_mask = np.zeros(frame.shape[0:2], dtype=np.uint8)
    for points in boxes:
        
        color = 255 # grayscale
        thickness = -1 # -1 for filled
        cv2.rectangle(exemplars_mask, tuple(points[0:2]), tuple(points[2:]),
                      color, thickness)    
     
    # get the skin mask    
    skin_mask = skin_detector.DetectSkinRegions(frame)
#    skin_mask_display = cv2.bitwise_and(frame, frame, mask=skin_mask)
    
    # get the magintude of optical flow
    mag = cv2.magnitude(flow[:, :, 0], flow[:, :, 1])
    
    # first box plot, take the upper whisker
    bp1 = plt.boxplot(np.reshape(mag, -1))
    plt.close()
    flow_mask1 = mag > bp1['whiskers'][1].get_data()[1][1]
    
    # second box plot on the uper whisker, threshold with the median
    bp2 = plt.boxplot(mag[flow_mask1])
    plt.close()
    flow_mask = \
        (mag > bp2['medians'][0].get_data()[1][0]).astype(np.uint8) * 255
    
#    # arrows
#    frame_gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
#    scaleFactor = 0.7
#    arrows = fns.draw_flow(frame_gray, flow * (1 / scaleFactor), flow_mask,
#                           step=4)
    
#    # overlay masked motion magnitude
#    connectivity = 3
#    cc_flow_mask = cv2.connectedComponentsWithStats(flow_mask, connectivity,
#                                                    cv2.CV_32S)
#    overlay_motion = overlay.overlay(frame_gray, cc_flow_mask,
#                                     np.unique(cc_flow_mask[1])[1:],
#                                     (255, 0, 0))
    
    # anding
    and_mask = cv2.bitwise_and(skin_mask, flow_mask)
#    fancy_and = cv2.bitwise_and(frame, frame, mask=and_mask)
    
    
#    # thresholded motion magnitude
#    motion_magnitude = cv2.bitwise_and(mag, mag, mask=flow_mask)
    
    
    # distance transform from fancy-and mask
    dt_fancy_and = cv2.distanceTransform(cv2.bitwise_not(and_mask),
                                         distanceType=cv2.DIST_L2,
                                         maskSize=cv2.DIST_MASK_PRECISE)
    dt_fancy_and_flow_mask = cv2.bitwise_and(dt_fancy_and, dt_fancy_and,
                                             mask=flow_mask)
    
    # grow around fancy and
    dt_thresh = 14.5
    
    # threshold using the distance from the fancy and mask
    dt_thresh, from_fancy_keep = cv2.threshold(dt_fancy_and_flow_mask,
                                               dt_thresh, 255,
                                               cv2.THRESH_BINARY_INV)
    from_fancy_keep = from_fancy_keep.astype(np.uint8)
    from_fancy_keep = cv2.bitwise_or(from_fancy_keep, and_mask,
                                     mask=flow_mask)  
    
    
    # separating into keep and toss (WAY #1 for components)
    connectivity = 3
    cc_dt_flow_keep = cv2.connectedComponentsWithStats(from_fancy_keep,
                                                       connectivity,
                                                       cv2.CV_32S)
    cc_dt_flow_keep = list(cc_dt_flow_keep)
    cc_dt_flow_keep.append([-1])
    
    # put -1 for background component so later appended labels line up
    keep_labels = [-1]
    toss_labels = [-1]
    
    for label in np.unique(cc_dt_flow_keep[1])[1:]:
        
        component_mask = np.zeros_like(exemplars_mask)
        component_mask[cc_dt_flow_keep[1] == label] = 255
        overlap = np.count_nonzero(cv2.bitwise_and(component_mask,
                                                   exemplars_mask))
        
        if overlap >= 0.2 * cc_dt_flow_keep[2][label, cv2.CC_STAT_AREA]:
            
            # keep
            cc_dt_flow_keep[-1].append(1)
            keep_labels.append(label)
            
        else:
            
            # toss
            cc_dt_flow_keep[-1].append(0)
            toss_labels.append(label)

    # tuple for WAY #1
    way_1_tuple = (cc_dt_flow_keep, keep_labels, toss_labels)
    


    '''WAY #2 for components (watershed lines used to break components)'''
    
    # distance transform on from_fancy_keep
    dt_from_fancy_keep = cv2.distanceTransform(from_fancy_keep,
                                               distanceType=cv2.DIST_L2,
                                               maskSize=cv2.DIST_MASK_PRECISE)
       
    # get rid of small connectors
    th = 0.2 * np.max(dt_from_fancy_keep)
    th, less_small_connectors = cv2.threshold(dt_from_fancy_keep, th, 255,
                                              cv2.THRESH_TOZERO)
    
    # per component, take the top 0.7
    markers_inner_mask = np.zeros(less_small_connectors.shape,
                                  dtype=np.uint8)
    markers_inner_mask[less_small_connectors != 0] = 255
    
    cc_dt = cv2.connectedComponents(markers_inner_mask)
    
    per_cc_top_percent_mask = np.zeros_like(less_small_connectors)
    
    for label in np.unique(cc_dt[1])[1:]:
        
        tmp = cv2.bitwise_and(less_small_connectors, less_small_connectors,
                              mask=(cc_dt[1] == label).astype(np.uint8))
        
        th_from_peak = np.max(tmp)
        
        per_cc_top_percent_mask[tmp > 0.75 * th_from_peak] = 255
    
    per_cc_top_percent_mask = per_cc_top_percent_mask.astype(np.uint8)
                    
    # local maximum
    local_max_mask = comp_fns.is_max_in_neighborhood(dt_from_fancy_keep, 6)
    
    # OR the two masks to get the inner mask for the marker
    inner_mask = cv2.bitwise_or(per_cc_top_percent_mask, local_max_mask)
    
    # dilate
    dilate_from_fancy_keep = cv2.dilate(from_fancy_keep, np.ones((43, 43)))
    
    # unknown
    unknown = cv2.subtract(dilate_from_fancy_keep, inner_mask)
    
    # connected components inner_mask
    cc_inner = cv2.connectedComponents(inner_mask)
    
    # markers
    markers = cc_inner[1] + 1
    markers[unknown == 255] = -1
    
    # watershed
#    markers_orig = np.copy(markers)
    markers = cv2.watershed(frame, markers)
    
    # marker lines (0) mask followed by erosion to separate
    marker_lines_mask = np.ones_like(markers)
    marker_lines_mask[markers == -1] = 0
    kernel = np.array([[0, 1, 0], [1, 1, 1], [0, 1, 0]])
    marker_lines_mask = cv2.morphologyEx(marker_lines_mask.astype(np.uint8),
                                         cv2.MORPH_ERODE, kernel)
    
    
    # break up the from_fancy_keep components with watershed (WS) lines
    broken_cc_label_matrix = (marker_lines_mask * \
                              cc_dt_flow_keep[1]).astype(np.uint8)
    broken_cc_label_matrix = comp_fns.fill_holes(broken_cc_label_matrix)
    
    # separating into keep and toss (WAY #2 for components)
    connectivity = 4
    cc_after_broken_cc = \
        cv2.connectedComponentsWithStats(broken_cc_label_matrix,
                                         connectivity, cv2.CV_32S)
    cc_after_broken_cc = list(cc_after_broken_cc)
    cc_after_broken_cc.append([-1])
    
    # put -1 for background component so later appended labels line up
    keep_labels = [-1]
    toss_labels = [-1]
    
    for label in np.unique(cc_after_broken_cc[1])[1:]:
        
        component_mask = np.zeros_like(exemplars_mask)
        component_mask[cc_after_broken_cc[1] == label] = 255
        overlap = np.count_nonzero(cv2.bitwise_and(component_mask,
                                                   exemplars_mask))
        
        if overlap >= 0.2 * cc_after_broken_cc[2][label, cv2.CC_STAT_AREA]:
            
            # keep
            cc_after_broken_cc[-1].append(1)
            keep_labels.append(label)
            
        else:
            
            # toss
            cc_after_broken_cc[-1].append(0)
            toss_labels.append(label)

    # tuple for WAY #2
    way_2_tuple = (cc_after_broken_cc, keep_labels, toss_labels)



    '''WAY #3 for components (WS components that overlap from_fancy_keep)'''
    
    # separating into keep and toss (WAY #3 for components)
    connectivity = 4
    cc_marker_lines_mask = cv2.connectedComponentsWithStats(marker_lines_mask,
                                                            connectivity,
                                                            cv2.CV_32S)
    # find the component with the maximum area and make it part of the background
    areas = cc_marker_lines_mask[2][:, cv2.CC_STAT_AREA]
    max_area = np.max(areas)
    max_area_index = np.where(areas == max_area)[0]
    
    small_WS_comp_mask = np.copy(marker_lines_mask)
    small_WS_comp_mask[cc_marker_lines_mask[1] == max_area_index] = 0
    
    # refind the components
    cc_small_WS_comp_mask = \
        cv2.connectedComponentsWithStats(small_WS_comp_mask, connectivity,
                                         cv2.CV_32S)
        
    # use components that overlap from_fancy_keep
    discard = []
    for label in np.unique(cc_small_WS_comp_mask[1][1:]):
        
        component_mask = np.zeros_like(from_fancy_keep)
        component_mask[cc_small_WS_comp_mask[1] == label] = 255
        overlap = np.count_nonzero(cv2.bitwise_and(component_mask,
                                                   from_fancy_keep))
        
        if overlap == 0:
        #if overlap <  0.25 * cc_small_WS_comp_mask[2][label,cv2.CC_STAT_AREA]:
            discard.append(component_mask)
    
    discard_m = np.any(discard, axis = 0)
    
    label_matrix_reduced = \
        np.copy(cc_small_WS_comp_mask[1]).astype(np.bool).astype(np.uint8)
    
    if len(discard_m.shape) != 0:
        
        label_matrix_reduced[discard_m] = 0
        
        # refind the cc if the label matrix reduced
        cc_small_WS_comp_mask = \
            cv2.connectedComponentsWithStats(label_matrix_reduced,
                                             connectivity, cv2.CV_32S)   
            
    cc_small_WS_comp_mask = list(cc_small_WS_comp_mask)
    cc_small_WS_comp_mask.append([-1])
    
    # put -1 for background component so later appended labels line up
    keep_labels = [-1]
    toss_labels = [-1]
    
    for label in np.unique(cc_small_WS_comp_mask[1])[1:]:
        
        component_mask = np.zeros_like(exemplars_mask)
        component_mask[cc_small_WS_comp_mask[1] == label] = 255
        overlap = np.count_nonzero(cv2.bitwise_and(component_mask,
                                                   exemplars_mask))
        
        if overlap >= 0.2 * cc_small_WS_comp_mask[2][label, cv2.CC_STAT_AREA]:
            
            # keep
            cc_small_WS_comp_mask[-1].append(1)
            keep_labels.append(label)
            
        else:
            
            # toss
            cc_small_WS_comp_mask[-1].append(0)
            toss_labels.append(label)

    # tuple for WAY #3
    way_3_tuple = (cc_small_WS_comp_mask, keep_labels, toss_labels)
    
    return ((way_1_tuple, way_2_tuple, way_3_tuple), flow_mask)






if __name__ == '__main__':
    
    '''
    path to pickled the resized color frame, its flow, and exemplar(s) information
    '''
    path_specific = 'STI Polk Mar30 F/MVI_0025_trans/'
    video_name = 'MVI_0025_trans__250_280.mp4'
    target_frame = 24
    
    path_common = './video_copies_clips_track/copies_and_clips_wave_02/'
    
    
    # pickle
    pickle_path = path_common + path_specific + 'frame_flow_exemplars__' + \
                  video_name[0:-4] + '__' + str(target_frame) + '.data'
                
    with open(pickle_path, 'rb') as f:
        frame, flow, boxes = pickle.load(f)
        
    # get gray frame
    frame_gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        
        
    # get components
    ways, flow_mask = get_cc_info(frame, flow, boxes)
    way_1, way_2, way_3 = ways[0], ways[1], ways[2]   
    
    # overlay (after separating into keep and toss) (WAY #1 for components)
    cc_dt_flow_keep, keep_labels, toss_labels = way_1
    
    overlay_motion_keep = overlay.overlay(frame_gray, cc_dt_flow_keep,
                                          keep_labels, (0, 255, 0), alpha=0.5)
    
    overlay_motion_keep = overlay.overlay(overlay_motion_keep, cc_dt_flow_keep,
                                          toss_labels, (0, 0, 255), alpha=0.5)
    
    
    # overlay (after separating into keep and toss) (WAY #2 for components)
    cc_after_broken_cc, keep_labels, toss_labels = way_2
    
    overlay_broken_by_WS = overlay.overlay(frame_gray, cc_after_broken_cc,
                                           keep_labels, (0, 255, 0), alpha=0.5)
    
    overlay_broken_by_WS = overlay.overlay(overlay_broken_by_WS,
                                           cc_after_broken_cc, toss_labels,
                                           (0, 0, 255), alpha=0.5)
    
    
    #  overlay (after separating into keep and toss) (WAY #3 for components)
    cc_small_WS_comp_mask, keep_labels, toss_labels = way_3
    
    overlay_WS_comp_kept = overlay.overlay(frame_gray,
                                           cc_small_WS_comp_mask, keep_labels,
                                           (0, 255, 0), alpha=0.5)
    
    overlay_WS_comp_kept = overlay.overlay(overlay_WS_comp_kept,
                                           cc_small_WS_comp_mask, toss_labels,
                                           (0, 0, 255), alpha=0.5)
        
        
    
        
    
    
    
    
    
    # display
    titles = ['components thus far',
              
              'components broken by watershed',
              
              'watershed components kept'
    
              ]
    
    images = [overlay_motion_keep,
              
              overlay_broken_by_WS,
              
              overlay_WS_comp_kept
    
             ]
    
            
    for title, image in zip(titles, images):
        
        cv2.namedWindow(title, cv2.WINDOW_GUI_NORMAL | cv2.WINDOW_AUTOSIZE)
        cv2.moveWindow(title, 0, -60) 
        cv2.imshow(title, image)
    
    k = cv2.waitKey(0)
            
    # escape
    if k == 27:
        
        escape = True
        
        cv2.destroyAllWindows()


























