# -*- coding: utf-8 -*-
"""
Created on Fri May 18 15:04:56 2018

@author: Callie Darsey

gather_features_01_wheeler
"""





import cv2
import numpy as np
import pickle
import matplotlib.pyplot as plt
import os


def force_neg_pi_to_pi(value):
    '''
    This function takes the value and adds or subtracts pi until a value in
    [-pi, pi] is reached and returned.
    '''
    while value > np.pi or value < -np.pi:
        
        if value > np.pi:
            
            value = value - 2 * np.pi
            
        elif value < -np.pi:
            
            value = value + 2 * np.pi
            
    return value


def get_bin_edges(start, bin_size):
    '''
    Return flow phase bins [start, start + 1 * bin_size),
    [start + 2 * bin_size), ..., [start + n * size, start) adjusted such that
    the bins capture angles around the unit circle taking values [-pi, pi].
    
    For example, if start = -np.pi / 8 and bin_size = np.pi / 4, the bins are
    
    [    -pi / 8,      pi / 8),
    [     pi / 8,  3 * pi / 8),
    [ 3 * pi / 8,  5 * pi / 8),
    [ 5 * pi / 8,  7 * pi / 8),
    [ 7 * pi / 8, -7 * pi / 8),
    [-7 * pi / 8, -5 * pi / 8),
    [-5 * pi / 8, -3 * pi / 8),
    [-3 * pi / 8,     -pi / 8)
    
    start - angle in radians to start the phase bins at
    bin_size - angle in radians of the bin widths    
    '''
    assert bin_size <= 2 * np.pi and bin_size > 0
    
    # set up the bin edges
    bin_edges = [start]
    
    angles_travelled = 0
    while True:
         
        angles_travelled = angles_travelled + bin_size
        
        if angles_travelled >= 2 * np.pi:
            
            bin_edges.append(start)
            
            break
        
        else:

            new_edge = force_neg_pi_to_pi(bin_edges[-1] + bin_size)
            
            bin_edges.append(new_edge)
    
    return bin_edges
    

def get_neg_pi_to_pi(values):
    '''
    This function takes the values and adds or subtracts pi until a value in
    [-pi, pi] is reached and returned.
    '''
    values_shape = values.shape
    values = values.ravel()
    
    indices_greater = np.where(values > np.pi)[0]
    indices_lesser = np.where(values < -np.pi)[0]
    
    while len(indices_greater) > 0 or len(indices_lesser) > 0:
        
        if len(indices_greater) > 0:
            
            values[indices_greater] = values[indices_greater] - 2 * np.pi
            
        if len(indices_lesser) > 0:
            
            values[indices_lesser] = values[indices_lesser] + 2 * np.pi
            
        indices_greater = np.where(values > np.pi)[0]
        indices_lesser = np.where(values < -np.pi)[0]
        
    values = np.reshape(values, values_shape)
            
    return values



def get_bin_associations(angles, bin_edges):
    '''
    angles - (rows, cols, frames) optical flow angles between -pi and pi
    bin_edges - bin edges generated from get_bin_edges(start, bin_size)
    '''
    
    # assert
    assert np.all([(angles < np.pi), (angles > -np.pi)])
    
    # find the associated bin indies for angles
    angles_bin_indices = np.zeros(angles.shape, dtype=np.uint)
    

    for  i, edge_pair in enumerate(zip(bin_edges[0:-1], bin_edges[1:])):
        
        edge_inclusive, edge_exclusive = edge_pair
        
        # The bin does not cross pi on unit circle
        if edge_inclusive < edge_exclusive:
            
            rows, cols, frames = np.where(np.all([edge_inclusive <= angles,
                                                  angles < edge_exclusive],
                                                axis=0))
            if len(rows) > 0 and len(cols) > 0 and len(frames) > 0:
                angles_bin_indices[rows, cols, frames] = i

                
        # The bin crosses pi on the unit circle
        elif edge_inclusive > edge_exclusive:
            
            rows, cols, frames = \
                np.where(np.any([np.all([edge_inclusive <= angles,
                                         angles <= np.pi], axis=0),
                                 np.all([-np.pi <= angles,
                                         angles < edge_exclusive], axis=0)],
                                axis=0))    
            
            if len(rows) > 0 and len(cols) > 0 and len(frames) > 0:
            
                angles_bin_indices[rows, cols, frames] = i
                
    return angles_bin_indices


def get_histograms_of_optical_flow(num_bins, flow_mag_block, bin_indices_block,
                                   col_1, col_2, row_1, row_2, segments):
    '''
    '''
    
    assert flow_mag_block.shape == bin_indices_block.shape
    
    flow_mag_block_patch = flow_mag_block[row_1 : row_2 + 1,
                                          col_1 : col_2 + 1, :]
    
    bin_indices_block_patch = bin_indices_block[row_1 : row_2 + 1,
                                                col_1 : col_2 + 1, :]
    
    feature_hists_of_flow                = np.zeros((segments, num_bins))
    feature_hists_of_flow_div_counts     = np.zeros((segments, num_bins))
    feature_hist_of_flow                 = np.zeros((1, num_bins))
    feature_hist_of_flow_div_counts      = np.zeros((1, num_bins))
    
    for i in range(num_bins):
        
        rows, cols, frames = np.where(bin_indices_block_patch == i)
                
        if len(frames) > 0:
            
            feature_hist_of_flow[0, i] = \
                np.sum(flow_mag_block_patch[rows, cols, frames])
            
            feature_hist_of_flow_div_counts[0, i] = \
                feature_hist_of_flow[0, i] / len(frames)
                
            
            for uf in np.unique(frames):
                
                ind = np.where(frames == uf)[0]
                
                r = rows[ind]
                c = cols[ind]
                f = frames[ind]
                
                feature_hists_of_flow[f, i] = \
                    np.sum(flow_mag_block_patch[r, c, f])
                    
                feature_hists_of_flow_div_counts[f, i] = \
                    feature_hists_of_flow[f, i] / len(f)
                    
    feature_sum_hists_of_flow_div_counts = \
        np.sum(feature_hists_of_flow_div_counts, axis=0)
    
    return (feature_hists_of_flow.ravel(),
            feature_hists_of_flow_div_counts.ravel(),
            feature_hist_of_flow,
            feature_sum_hists_of_flow_div_counts,
            feature_hist_of_flow_div_counts)


def combine(crop1, crop2):
    '''
    '''
    mask1 = np.any(crop1 > 50, axis=2)
    mask2 = np.any(crop2 > 50, axis=2)
    
    intersection = np.logical_and(mask1, mask2)
    xor_mask1 = np.logical_xor(mask1, intersection)
    xor_mask2 = np.logical_xor(mask2, intersection)    

    pred_side_1 = np.copy(crop1)
    pred_side_2 = np.copy(crop2)
    
    np.copyto(pred_side_1, crop2,
              where=np.stack([xor_mask2, xor_mask2, xor_mask2], axis=2))
    np.copyto(pred_side_2, crop1,
              where=np.stack([xor_mask1, xor_mask1, xor_mask1], axis=2))

    alpha = 0.5
    beta = 1.0 - alpha
    gamma = 0.0
    
    output = cv2.addWeighted(pred_side_1, alpha, pred_side_2, beta, gamma)
    
    return output


def get_seg_skin_overlays(segment_cleaned_skin, col_1, col_2, row_1, row_2):
    '''
    '''
    
    crop1 = segment_cleaned_skin[0][row_1 : row_2 + 1, col_1 : col_2 + 1, :]
    
    for index in np.arange(1, len(segment_cleaned_skin)):
        
        crop2 = segment_cleaned_skin[index][row_1 : row_2 + 1,
                                            col_1 : col_2 + 1, :]

        crop1 = combine(crop1, crop2)
        
    return crop1




def function(bin_size=32):
    '''
    '''
   
    e1 = cv2.getTickCount()
    
    directory = "./"
    
    for content in os.listdir(directory):
        
        if os.path.isfile(directory + content):
            
            if content[-12:] == '.sampled.mp4':
                
                name = content[:-12]
                
                print(name)
                
                break
                                       
    cap_flow_mag = cv2.VideoCapture(directory+'flow_'+name+'.flow_mag.mp4')
    cap_flow_ang = cv2.VideoCapture(directory+'flow_'+name+'.flow_ang.mp4')
    cap_cleaned_skin = cv2.VideoCapture(directory+name+'.cleaned_skin.mp4')
    cap_frame = cv2.VideoCapture(directory+name+'.sampled.mp4')   
    
    
    with open(directory + 'patches_all_segments_labels___' + name + '.data',
              'rb') as f:
        
        patches_and_labels = pickle.load(f)
        
    keys = np.array(list(patches_and_labels.keys()))
    
    
    patches_and_labels_and_feature_watershed_skin_bool = {}
    patches_and_labels_and_feature_skin_count = {}
    patches_and_labels_and_feature_skin_bool = {}
    patches_and_labels_and_feature_hists_of_flow = {}
    patches_and_labels_and_feature_hists_of_flow_div_counts = {}
    patches_and_labels_and_feature_hist_of_flow = {}
    patches_and_labels_and_feature_sum_hists_of_flow_div_counts = {}
    patches_and_labels_and_feature_hist_of_flow_div_counts = {}
    
    fdicts = [patches_and_labels_and_feature_watershed_skin_bool,
              patches_and_labels_and_feature_skin_count,
              patches_and_labels_and_feature_skin_bool,
              patches_and_labels_and_feature_hists_of_flow,
              patches_and_labels_and_feature_hists_of_flow_div_counts,
              patches_and_labels_and_feature_hist_of_flow,
              patches_and_labels_and_feature_sum_hists_of_flow_div_counts,
              patches_and_labels_and_feature_hist_of_flow_div_counts]
    
    ftags = ['feature_watershed_skin_bool',
             'feature_skin_count',
             'feature_skin_bool',
             'feature_correct_hists_of_flow',
             'feature_correct_hists_of_flow_div_counts',
             'feature_correct_hist_of_flow',
             'feature_correct_sum_hists_of_flow_div_counts',
             'feature_correct_hist_of_flow_div_counts']    


    segment = 45
    
    segment_flow_mag = []
    segment_flow_ang = []
    segment_cleaned_skin = []
    segment_cleaned_skin_watershed = []
    segment_frame = []
    
    count = -1
    ind = -1
    
    while(cap_flow_mag.isOpened() and cap_flow_ang.isOpened() and \
          cap_cleaned_skin.isOpened() and cap_frame.isOpened()):
        
        success_flag_flow_mag, flow_mag_display    = cap_flow_mag.read(0)
        success_flag_flow_ang, flow_ang_display     = cap_flow_ang.read(0)
        success_flag_skin,     cleaned_skin_display = cap_cleaned_skin.read()
        success_frame,         frame_display        = cap_frame.read()
        
        if success_flag_flow_mag == True and success_flag_flow_ang == True \
            and success_flag_skin == True and success_frame == True:
            
            count = count + 1
            
            rows, cols, channels = frame_display.shape
            
            
            # watershed skin
            skin_mask_thr = np.any(cleaned_skin_display > 50,
                                   axis=2).astype(np.uint8)
    
            
            dilate_skin_mask_thr = cv2.dilate(skin_mask_thr,
                                              kernel=np.ones((3,3)),
                                              iterations=80)
            
            unknown_skin_watershed = cv2.subtract(dilate_skin_mask_thr,
                                                  skin_mask_thr)
            
            cc_skin_mask_thr = cv2.connectedComponents(skin_mask_thr)
            
            markers_skin_watershed = cc_skin_mask_thr[1] + 1
            
            markers_skin_watershed[unknown_skin_watershed > 0] = -1
            
            # call the watershed
            cv2.watershed(frame_display, markers_skin_watershed)
            
            cc_markers_skin_watershed = cv2.connectedComponentsWithStats(\
                markers_skin_watershed.astype(np.bool).astype(np.uint8))
                
            skin_watershed = markers_skin_watershed.copy()
            
            skin_watershed[skin_watershed == -1] = 0
            skin_watershed[skin_watershed == np.argmax(\
                cc_markers_skin_watershed[2][:, cv2.CC_STAT_AREA])] = 0
    
           
            segment_flow_mag.append(flow_mag_display[:, :, 0])
            segment_flow_ang.append(flow_ang_display[:, :, 0])
            segment_cleaned_skin.append(cleaned_skin_display)
            segment_cleaned_skin_watershed.append(skin_watershed)
            segment_frame.append(frame_display)
    
        
            if len(segment_flow_mag) > segment:
                segment_flow_mag.pop(0)
                segment_flow_ang.pop(0)
                segment_cleaned_skin.pop(0)
                segment_cleaned_skin_watershed.pop(0)
                segment_frame.pop(0)
    
            else:
                ind = ind + 1
    
    
            if count != 0 and (count + 1) % segment == 0:
                
                ei = cv2.getTickCount()
                
                print('segment', int(np.floor(count / segment)))
    
                # find if segment has exemplars to collect features for
                index = np.where(keys[:, 1] == count * 4)[0]
    
                if len(index) == 1:
                    
                    # key for the segment
                    relevant_key = (keys[index[0], 0], keys[index[0], 1])
                    
                    # patch box info (and tags) for the segment
                    patches, tags = patches_and_labels[relevant_key]
    
                    # get the patch size
                    col_1, col_2 = np.sort([patches[0][0], patches[0][2]])
                
                    row_1, row_2 = np.sort([patches[0][1], patches[0][3]])
                    
    
                    '''operations on the  full frames over the segment'''
                    
                    # skin watershed union over time
                    skin_watershed_union = \
                        np.any(segment_cleaned_skin_watershed, axis=0)
        
                    # overlay the cleaned skin
                    skin_cleaned_combined = get_seg_skin_overlays(\
                        segment_cleaned_skin, 0, cols, 0, rows)
                    
                    # grayscale of the overlay cleaned skin
                    skin_cleaned_combined_gray = \
                        cv2.cvtColor(skin_cleaned_combined, cv2.COLOR_BGR2GRAY)
                        
                    # for histogrom of flow
                    
                    flow_mag_display_block = np.stack(segment_flow_mag, axis=2)
                    
                    flow_ang_radian_block = np.stack(segment_flow_ang, axis=2) * \
                                            (360.0 / 255.0) * (np.pi / 180.0)
                                            
                    flow_ang_radian_block = get_neg_pi_to_pi(flow_ang_radian_block)
                    
                    bin_edges = get_bin_edges(-2 * np.pi / bin_size / 2,
                                               2 * np.pi / bin_size)
                    
                    bin_indices = get_bin_associations(flow_ang_radian_block,
                                                       bin_edges)
                    
                    num_bins = len(bin_edges) - 1
    

    
                    '''
                    ***********************
                    features to collect
                    ***********************
                    '''
                    
                    ''' watershed skin bool '''                 
                    feature_watershed_skin_bool = np.zeros((patches.shape[0], 1))
                                        
                    ''' number of nonzero elements in color overlay skin image '''
                    feature_skin_count = np.zeros((patches.shape[0], 1))
                    
                    ''' skin overlay bool'''
                    feature_skin_bool = np.zeros((patches.shape[0], 1))
                    
                    ''' hist of flow per patch per frame '''
                    feature_hists_of_flow = np.zeros((patches.shape[0],
                                                      bin_size * segment))
                    
                    ''' hist of flow per patch per frame divided by count '''
                    feature_hists_of_flow_div_counts = np.zeros((patches.shape[0],
                                                            bin_size * segment))
                    
                    ''' hist of flow per patch through time '''
                    feature_hist_of_flow = np.zeros((patches.shape[0], bin_size))
                    
                    ''' sum hist of flow per patch per frame divided by count '''
                    feature_sum_hists_of_flow_div_counts = \
                        np.zeros((patches.shape[0], bin_size))
                    
                    ''' hist of flow per patch through time divided by count '''
                    feature_hist_of_flow_div_counts = np.zeros((patches.shape[0],
                                                                bin_size))                    
                    
    
                    # loop through the patches
                    for i, patch in enumerate(patches):
                        
                        col_1, col_2 = np.sort([patch[0], patch[2]])
                    
                        row_1, row_2 = np.sort([patch[1], patch[3]])
                        
    
                        feature_watershed_skin_bool[i] = int(np.sum(\
                            skin_watershed_union[row_1 : row_2 + 1,
                                                 col_1 : col_2 + 1]) > 0)
                                                
                        feature_skin_count[i] = np.count_nonzero(\
                            skin_cleaned_combined_gray[row_1 : row_2 + 1,
                                                       col_1 : col_2 + 1])
                        
                        feature_skin_bool[i] = int(np.count_nonzero(\
                            feature_skin_count[i] > 0))

                        patch_feature_hists_of_flow, \
                        patch_feature_hists_of_flow_div_counts, \
                        patch_feature_hist_of_flow, \
                        patch_feature_sum_hists_of_flow_div_counts, \
                        patch_feature_hist_of_flow_div_counts = \
                            get_histograms_of_optical_flow(num_bins,
                                                           flow_mag_display_block,
                                                           bin_indices,
                                                           col_1, col_2,
                                                           row_1, row_2, segment)
                            
                        feature_hists_of_flow[i, :] = patch_feature_hists_of_flow
    
                        feature_hists_of_flow_div_counts[i, :] = \
                            patch_feature_hists_of_flow_div_counts
    
                        feature_hist_of_flow[i, :] = patch_feature_hist_of_flow
    
                        feature_sum_hists_of_flow_div_counts[i, :] = \
                            patch_feature_sum_hists_of_flow_div_counts
                        
                        ''' hist of flow per patch through time divided by count '''
                        feature_hist_of_flow_div_counts[i, :] = \
                            patch_feature_hist_of_flow_div_counts   

                        
    
                    # put into the dictionaries
                    info_and_features = [feature_watershed_skin_bool,
                                         feature_skin_count,
                                         feature_skin_bool,
                                         feature_hists_of_flow,
                                         feature_hists_of_flow_div_counts,
                                         feature_hist_of_flow,
                                         feature_sum_hists_of_flow_div_counts,
                                         feature_hist_of_flow_div_counts]
                
                    
                    for info_or_feature, fdict in zip(info_and_features, fdicts):
                        
                        if (relevant_key) not in fdict.keys():
                            fdict[(relevant_key)] = []
                            fdict[(relevant_key)].append(patches)
                            fdict[(relevant_key)].append(tags)
                        
                        fdict[(relevant_key)].append(info_or_feature)
    
                eb = cv2.getTickCount()
                time = (eb - ei)/ cv2.getTickFrequency()
                print("Segment execution time in seconds = ", time)    
              
        else:
            
            print("No more frames")
            
            cap_flow_mag.release()
            cap_flow_ang.release()
            cap_cleaned_skin.release()
            cap_frame.release()  
                   
            break        
    else:
        
        cap_flow_mag.release()
        cap_flow_ang.release()
        cap_cleaned_skin.release()
        cap_frame.release()     
    
    
    for ftag, fdict in zip(ftags, fdicts):
    
        with open(directory + 'patches_all_segments_labels___' + ftag + \
                  '_Bins_{:d}'.format(bin_size) + \
                  '___' + name + '.data', 'wb') as f:
        
            pickle.dump((fdict), f)
        
    e2 = cv2.getTickCount()
    time = (e2 - e1)/ cv2.getTickFrequency()
    print("Total execution time in seconds = ", time)
