# -*- coding: utf-8 -*-
"""
Created on Mon Mar 26 11:29:06 2018

@author: Callie Darsey

get lessened features and save the steps


Inputs:
    color frame
    flow for frame
    
Outputs:
    intermediate images for frame
        flow thresholded
        skin detection
        moving skin mask
        distance transform on moving skin mask
        way 1 - threshold of distance transform on moving skin mask
        way 2 - use watershed lines to break up
        way 3 - use watershed components overlapping way 1
    labelled connected components and eaches features vector

get_lessened_features_and_save_steps_01

"""

import numpy as np
import cv2
import skin_detector_Abby_01 as skin_detector
from matplotlib import pyplot as plt
import component_functions_01 as comp_fns



# column indices into features
FEATURE_H_PDF_LEFT = 0
FEATURE_H_PDF_RIGHT = 17
FEATURE_S_PDF_LEFT = 18
FEATURE_S_PDF_RIGHT = 33
FEATURE_V_PDF_LEFT = 34
FEATURE_V_PDF_RIGHT = 49

FEATURE_FLOW_PHASE_BIN_MAG_SUM_HIST_DIV_COUNT_LEFT = 50
FEATURE_FLOW_PHASE_BIN_MAG_SUM_HIST_DIV_COUNT_RIGHT = 57

FEATURE_CANNY_EDGE_TO_AREA_RATIO = 58

FEATURE_AREA = 59
FEATURE_PERIMETER = 60

FEATURE_MIN_AREA_RECT_ORIENT = 61
FEATURE_AREA_TO_MIN_AREA_RECT = 62
FEATURE_MIN_AREA_RECT_SHORT_TO_LONG_SIDES_RATIO = 63

features_vector_size = FEATURE_MIN_AREA_RECT_SHORT_TO_LONG_SIDES_RATIO + 1

def is_max_in_neighborhood(image, radius):
    '''
    Determine if each pixel is the maximum in a neighborhood.
    
    image - one channel image
    radius - radius of the square used as a kernel
    '''
    
    with_border = cv2.copyMakeBorder(image, radius, radius,
                                     radius, radius, cv2.BORDER_CONSTANT,
                                     None, 0)
    
    collect = []
    
    for i in range(0, radius + 1):
        for j in range(0, radius + 1):
            
            if (i != 0) and (j != 0):
        
                collect.append(image > with_border[\
                                    radius + i : image.shape[0] + radius + i,
                                    radius + j : image.shape[1] + radius + j])
            
                collect.append(image > with_border[\
                                    radius - i : image.shape[0] + radius - i,
                                    radius - j : image.shape[1] + radius - j])
            
                collect.append(image > with_border[\
                                    radius - i : image.shape[0] + radius - i,
                                    radius + j : image.shape[1] + radius + j])
            
                collect.append(image > with_border[\
                                    radius + i : image.shape[0] + radius + i,
                                    radius - j : image.shape[1] + radius - j])

    mark = np.zeros(image.shape, dtype=np.uint8)
    
    mark[np.all(collect, axis=0)] = 255
    
    return mark

def force_neg_pi_to_pi(value):
    '''
    This function takes the value and adds or subtracts pi until a value in
    [-pi, pi] is reached and returned.
    '''
    while value > np.pi or value < -np.pi:
        
        if value > np.pi:
            
            value = value - 2 * np.pi
            
        elif value < -np.pi:
            
            value = value + 2 * np.pi
            
    return value


    
def get_flow_mag_in_phase_bins(flow, mask, start, bin_size):
    '''
    Return a histogram with flow phase bins [start, start + 1 * bin_size),
    [start + 2 * bin_size), ..., [start + n * size, start) adjusted such that
    the bins capture angles around the unit circle taking values [-pi, pi].
    
    For example, if start = -np.pi / 8 and bin_size = np.pi / 4, the bins are
    
    [    -pi / 8,      pi / 8),
    [     pi / 8,  3 * pi / 8),
    [ 3 * pi / 8,  5 * pi / 8),
    [ 5 * pi / 8,  7 * pi / 8),
    [ 7 * pi / 8, -7 * pi / 8),
    [-7 * pi / 8, -5 * pi / 8),
    [-5 * pi / 8, -3 * pi / 8),
    [-3 * pi / 8,     -pi / 8)
    
    Three versions of this histogram are returned.
    
    1) The value in each bin will be the sum of the flow magnitude values of
       pixels whose flow phase was within the phase bin range.
    
    2) The histogram described in 1) with each value divided by the phase bin
       count (if the count is nonzero)
       
    3) The histogram described in 1) with each value divided by the mask (i.e.
        component) area

    Inputs:
        
    flow - two channel array holding the bin sizes
    mask - mask of the connected component
    start - angle in radians to start the phase bins at
    bin_size - angle in radians of the bin widths

    '''
    assert bin_size <= 2 * np.pi and bin_size > 0
    
    # force start to be in [-pi, pi]
    start = force_neg_pi_to_pi(start)
    
    # set up the bin edges
    bin_edges = [start]
    
    angles_travelled = 0
    while True:
         
        angles_travelled = angles_travelled + bin_size
        
        if angles_travelled >= 2 * np.pi:
            
            bin_edges.append(start)
            
            break
        
        else:

            new_edge = force_neg_pi_to_pi(bin_edges[-1] + bin_size)
            
            bin_edges.append(new_edge)

    # get the area of the mask (i.e. the area of the component)
    area = np.count_nonzero(mask)

    # get the flow into magnitude and angle form
    flow_mag = cv2.magnitude(flow[:, :, 0], flow[:, :, 1])
    flow_phase = cv2.phase(flow[:, :, 0], flow[:, :, 1])

    # find the sums and counts for the bins
    
    bin_sums = np.zeros((1, len(bin_edges)-1), dtype=np.float32)
    bin_counts = np.zeros((1, len(bin_edges)-1), dtype=np.float32)
    
    r_ind, c_ind = np.where(mask)
    
    for r, c in zip(r_ind, c_ind):
        
        flow_phase_value = force_neg_pi_to_pi(flow_phase[r, c])      
        
        for  i, edge_pair in enumerate(zip(bin_edges[0:-1], bin_edges[1:])):
            
            edge_inclusive, edge_exclusive = edge_pair
            
            # The bin does not cross pi on unit circle
            if edge_inclusive < edge_exclusive:
                
                if edge_inclusive <= flow_phase_value and \
                                     flow_phase_value < edge_exclusive:
                                       
                    bin_sums[0, i] = bin_sums[0, i] + flow_mag[r, c]
                    
                    bin_counts[0, i] = bin_counts[0, i] + 1
                    
                    break
                    
            # The bin crosses pi on the unit circle
            elif edge_inclusive > edge_exclusive:
                
                if edge_inclusive <= flow_phase_value and \
                                     flow_phase_value <= np.pi or \
                           -np.pi <= flow_phase_value and \
                                     flow_phase_value < edge_exclusive:
                
                    bin_sums[0, i] = bin_sums[0, i] + flow_mag[r, c]
                    
                    bin_counts[0, i] = bin_counts[0, i] + 1
                    
                    break
    
    # divide sums by bin counts            
    bin_sums_div_by_count = np.copy(bin_sums)

    bin_sums_div_by_count[bin_counts != 0] = \
        bin_sums_div_by_count[bin_counts != 0] / bin_counts[bin_counts != 0]
        
    # divide sums by area
    bin_sums_div_by_area = bin_sums / area
    
    return bin_sums, bin_sums_div_by_count, bin_sums_div_by_area


def get_threshold_flow_mask(mag, lower_thresh=None):
    '''
    Threshold flow in the frame with the median of the flow magnitudes found
    above the higher whisker of all the flow magnitudes in the whole frame.
    '''
    
    # first box plot on magitudes
    bp1 = plt.boxplot(np.reshape(mag, -1))
    plt.close()
    
    # extract magnitudes above the upper whisker
    bp1_thresh = bp1['whiskers'][1].get_data()[1][1]
    mag_above_whisker = mag[mag > bp1_thresh]
    
    # second box plot for magnitudes above the upper whisker
    bp2 = plt.boxplot(mag_above_whisker.reshape((-1, 1)))
    plt.close()
    
    # use the lower of the second boxplot median and passed threshold
    bp2_thresh = bp2['medians'][0].get_data()[1][0]
    
    if lower_thresh is not None:
        
        flow_mag_thresh = np.min([lower_thresh, bp2_thresh])
        
    else:
        
        flow_mag_thresh = bp2_thresh
        
    flow_mask = (mag > bp2_thresh).astype(np.uint8) * 255
        
    return (flow_mask, bp1_thresh, bp2_thresh, flow_mag_thresh, bp1, bp2)


def get_components_and_intermediate_steps(frame, flow, lower_thresh,
                                          dt_thresh = 14.5):
    '''
    
    This function returns 
    
    The three ways of having componnents are as follows:
        
        * WAY #1 for components found via
              1) finding the intersection (AND) between the skin detection and
                 thresholded motion
              2) taking the distance transform (DT or dt) on the intersection
                 and thresholding the DT at some chosen threshold value
              3) finding the intersection (AND) of the DT thresholded and the
                 thresholded motion
                 
        * WAY #2 for components found via        
              1) find the components according to WAY #1
              2) find watershed lines using the distance transform on WAY #1
                 components to mark the watershed function marker input
              2) break the WAY #1 components with the dilated watershed
                 (WS or ws) marker lines
                 
        * WAY #3 for components found via
              1) use the dilated watershed markers lines of WAY #2 and reject
                 the largest watershed (background) component
              2) if a watershed component overlaps a component of WAY #1 at
                 all, use as component
    
    Each of the returned components has the cv2.connectedComponentsWithStats()
    output with an appended element holding the label 1 for 'keep' and 0 for
    'toss' and -1 for the background component with label 0. Furthermore, it
    has lists of the keep labels and toss labels.
    
    frame - BGR color image
    flow - two channel flow image

    '''
    
    '''
    collect for return so that intermediate images and values can be saved
    and viewed
    '''
    intermediate_tag = []
    intermediate_arr = []

        
    '''WAY #1 for components (no watershed)'''
     
    # get the skin mask    
    skin_mask = skin_detector.DetectSkinRegions(frame)
    
    # save for return
    skin_mask_display = cv2.bitwise_and(frame, frame, mask=skin_mask)
    intermediate_tag.append('skin_detection')
    intermediate_arr.append(skin_mask_display)  
    
    # get the magintude of optical flow
    mag = cv2.magnitude(flow[:, :, 0], flow[:, :, 1])
    
    # first box plot on magitudes
    bp1 = plt.boxplot(np.reshape(mag, -1))
    plt.close()
    
    # extract magnitudes above the upper whisker
    bp1_thresh = bp1['whiskers'][1].get_data()[1][1]
    mag_above_whisker = mag[mag > bp1_thresh]
    
    # second box plot for magnitudes above the upper whisker
    bp2 = plt.boxplot(mag_above_whisker.reshape((-1, 1)))
    plt.close()
    
    # use the lower of the second boxplot median and passed threshold
    bp2_thresh = bp2['medians'][0].get_data()[1][0]
    flow_mag_thresh = np.min([lower_thresh, bp2_thresh])

    flow_mask = (mag > flow_mag_thresh).astype(np.uint8) * 255
    
    
    # save for return
    intermediate_tag.append('flow_mask')
    intermediate_arr.append(flow_mask)         

    intermediate_tag.append('flow_mag_thresh')
    intermediate_arr.append(flow_mag_thresh)  
    
    # anding
    and_mask = cv2.bitwise_and(skin_mask, flow_mask)

    # save for return
    frame_with_and_mask = cv2.bitwise_and(frame, frame, mask=and_mask)
    intermediate_tag.append('skin_and_flow_mag_thresh')
    intermediate_arr.append(frame_with_and_mask)      
           
    # distance transform from skin and flow mag. thresh. mask (and_mask)
    dt_fancy_and = cv2.distanceTransform(cv2.bitwise_not(and_mask),
                                         distanceType=cv2.DIST_L2,
                                         maskSize=cv2.DIST_MASK_PRECISE)
    dt_fancy_and_flow_mask = cv2.bitwise_and(dt_fancy_and, dt_fancy_and,
                                             mask=flow_mask)
    
    # save for return
    intermediate_tag.append('DT_skin_and_flow_mag_thresh')
    intermediate_arr.append(dt_fancy_and)

    intermediate_tag.append('DT_skin_and_flow_mag_thresh_AND_flow_mag_thresh')
    intermediate_arr.append(dt_fancy_and_flow_mask)      
    
    # grow around skin and flow mag. thresh. mask (and_mask)
    # dt_thresh = 14.5
    
    # threshold using the distance from the fancy and mask
    dt_thresh, from_fancy_keep = cv2.threshold(dt_fancy_and_flow_mask,
                                               dt_thresh, 255,
                                               cv2.THRESH_BINARY_INV)
    from_fancy_keep = from_fancy_keep.astype(np.uint8)
    from_fancy_keep = cv2.bitwise_or(from_fancy_keep, and_mask,
                                     mask=flow_mask)
    
    # save for return
    intermediate_tag.append('way_1_components_mask')
    intermediate_arr.append(from_fancy_keep)          

    

    '''WAY #2 for components (watershed lines used to break components)'''
    
    # distance transform on from_fancy_keep
    dt_from_fancy_keep = cv2.distanceTransform(from_fancy_keep,
                                               distanceType=cv2.DIST_L2,
                                               maskSize=cv2.DIST_MASK_PRECISE)
    
    # save for return
    intermediate_tag.append('DT_from_way_1_components_mask')
    intermediate_arr.append(dt_from_fancy_keep)            
       
    # get rid of small connectors
    th = 0.2 * np.max(dt_from_fancy_keep)
    th, less_small_connectors = cv2.threshold(dt_from_fancy_keep, th, 255,
                                              cv2.THRESH_TOZERO)
    
    # save for return
    intermediate_tag.append('less_small_connectors')
    intermediate_arr.append(less_small_connectors)    
    
    # per component, take the top 0.75
    markers_inner_mask = np.zeros(less_small_connectors.shape,
                                  dtype=np.uint8)
    markers_inner_mask[less_small_connectors != 0] = 255
    
    cc_dt = cv2.connectedComponents(markers_inner_mask)
    
    per_cc_top_percent_mask = np.zeros_like(less_small_connectors)
    
    for label in np.unique(cc_dt[1])[1:]:
        
        tmp = cv2.bitwise_and(less_small_connectors, less_small_connectors,
                              mask=(cc_dt[1] == label).astype(np.uint8))
        
        th_from_peak = np.max(tmp)
        
        per_cc_top_percent_mask[tmp > 0.75 * th_from_peak] = 255
    
    per_cc_top_percent_mask = per_cc_top_percent_mask.astype(np.uint8)
                    
    # local maximum
    local_max_mask = comp_fns.is_max_in_neighborhood(dt_from_fancy_keep, 6)


    
    # save for return
    intermediate_tag.append('per_cc_top_percent_mask')
    intermediate_arr.append(per_cc_top_percent_mask)

    intermediate_tag.append('local_max_mask')
    intermediate_arr.append(local_max_mask)        
    
    # OR the two masks to get the inner mask for the marker
    inner_mask = cv2.bitwise_or(per_cc_top_percent_mask, local_max_mask)
              
    # dilate
    dilate_from_fancy_keep = cv2.dilate(from_fancy_keep, np.ones((43, 43)))
    
    # unknown
    unknown = cv2.subtract(dilate_from_fancy_keep, inner_mask)
    
    # connected components inner_mask
    cc_inner = cv2.connectedComponents(inner_mask)
    
    # markers
    markers = cc_inner[1] + 1
    markers[unknown == 255] = -1
    
    # save for return
    intermediate_tag.append('markers')
    intermediate_arr.append(np.copy(markers))          
    
    # watershed
#    markers_orig = np.copy(markers)
    markers = cv2.watershed(frame, markers)
    
    # save for return
    intermediate_tag.append('watershed_lines')
    intermediate_arr.append(markers)       
    
    # marker lines (0) mask followed by erosion to separate
    marker_lines_mask = np.ones_like(markers)
    marker_lines_mask[markers == -1] = 0
    kernel = np.array([[0, 1, 0], [1, 1, 1], [0, 1, 0]])
    marker_lines_mask = cv2.morphologyEx(marker_lines_mask.astype(np.uint8),
                                         cv2.MORPH_ERODE, kernel)
    
    # break up the from_fancy_keep components with watershed (WS) lines
    broken_cc_label_matrix = (marker_lines_mask * \
                              from_fancy_keep).astype(np.uint8)
    broken_cc_label_matrix = comp_fns.fill_holes(broken_cc_label_matrix)
    
    # save for return
    intermediate_tag.append('way_2_components_mask')
    intermediate_arr.append(broken_cc_label_matrix)    



    '''WAY #3 for components (WS components that overlap from_fancy_keep)'''
    
    # separating into keep and toss (WAY #3 for components)
    connectivity = 4
    cc_marker_lines_mask = cv2.connectedComponentsWithStats(marker_lines_mask,
                                                            connectivity,
                                                            cv2.CV_32S)
    # find the component with the maximum area and make it part of the background
    areas = cc_marker_lines_mask[2][:, cv2.CC_STAT_AREA]
    max_area = np.max(areas)
    max_area_index = np.where(areas == max_area)[0]
    
    small_WS_comp_mask = np.copy(marker_lines_mask)
    small_WS_comp_mask[cc_marker_lines_mask[1] == max_area_index] = 0
     
    # refind the components
    cc_small_WS_comp_mask = \
        cv2.connectedComponentsWithStats(small_WS_comp_mask, connectivity,
                                         cv2.CV_32S)
  
    # save for return
    intermediate_tag.append('watershed_components')
    intermediate_arr.append(cc_small_WS_comp_mask[1])      
     
        
    # use components that overlap from_fancy_keep
    discard = []
    for label in np.unique(cc_small_WS_comp_mask[1][1:]):
        
        component_mask = np.zeros_like(from_fancy_keep)
        component_mask[cc_small_WS_comp_mask[1] == label] = 255
        overlap = np.count_nonzero(cv2.bitwise_and(component_mask,
                                                   from_fancy_keep))
        
        if overlap == 0:
        #if overlap <  0.25 * cc_small_WS_comp_mask[2][label,cv2.CC_STAT_AREA]:
            discard.append(component_mask)
    
    discard_m = np.any(discard, axis = 0)
    
    label_matrix_reduced = \
        np.copy(cc_small_WS_comp_mask[1]).astype(np.bool).astype(np.uint8)
    
    if len(discard_m.shape) != 0:
        
        label_matrix_reduced[discard_m] = 0
        
        # refind the cc if the label matrix reduced
        cc_small_WS_comp_mask = \
            cv2.connectedComponentsWithStats(label_matrix_reduced,
                                             connectivity, cv2.CV_32S)
            
    # save for return
    intermediate_tag.append('way_3_components_mask')
    intermediate_arr.append(label_matrix_reduced)

    return (intermediate_tag, intermediate_arr)


def get_features(cc, frame, flow):
    '''
    cc - Output of cv2.connectedComponentsWithStats()          
    frame - resized color frame
    flow - two channel flow for resized color frame
    
    Return cc with another element holding the features vector. (The
    background is filled with -1 values.)
    '''
   
    # append another element to hold features (fill background with -1)
    cc = list(cc)
    cc.append([])
    cc[-1].append(-1 * np.ones((1, features_vector_size), dtype=np.float32))
    
    # convert the color array to HSV
    frame_HSV = cv2.cvtColor(frame, cv2.COLOR_BGR2HSV)    
    
    # prep to go through the channels of HSV and to get pdfs
    
    # bin edges
    bin_edges_0to255 = np.arange(0, 257, 16)
    bin_edges_0to180 = np.arange(0, 181, 10)
    bin_edges_arrays = [bin_edges_0to180] + \
                       [bin_edges_0to255 for i in range(2)]
                
    # indices: start inclusive, end exclusive

    start_end_pairs_pdf = [[FEATURE_H_PDF_LEFT, FEATURE_H_PDF_RIGHT + 1],
                           [FEATURE_S_PDF_LEFT, FEATURE_S_PDF_RIGHT + 1],
                           [FEATURE_V_PDF_LEFT, FEATURE_V_PDF_RIGHT + 1]]
    
    # canny
    '''https://docs.opencv.org/3.1.0/da/d22/tutorial_py_canny.html'''
    minVal = 50
    maxVal = 150
    Y = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
    edges = cv2.Canny(Y, minVal, maxVal)
    
    
    
    # now loop through the rest of the components and gather features
    for label in np.unique(cc[1])[1:]:
    
        features_vector = np.zeros((1, features_vector_size), dtype=np.float32)
            
        cc_bool_mask = (cc[1] == label)
        
        dilated_cc_mask = cv2.dilate(cc_bool_mask.astype(np.uint8),
                                     kernel=np.ones((3,3)))
        
        cc_dilated_cc = cv2.connectedComponentsWithStats(dilated_cc_mask)
        cc_dilated_bool_mask = cc_dilated_cc[1].astype(np.bool)
        cc_dilated_mask = cc_dilated_bool_mask.astype(np.uint8)
        
        # form the color space PDFs    
        for i, info in enumerate(zip(bin_edges_arrays, start_end_pairs_pdf)):
            
            bin_edges, start_end = info
    
            hist, bin_edges = np.histogram(frame_HSV[:, :, i] \
                                           [cc_dilated_bool_mask],
                                           bin_edges)
        
            features_vector[:, start_end[0]:start_end[1]] = hist / np.sum(hist)
               
            
        # get the canny edge to area ratio (using the dilated component area)
        edges_in_dilated_cc = cv2.bitwise_and(edges, edges,
                                              mask=cc_dilated_mask)
        
        features_vector[:, FEATURE_CANNY_EDGE_TO_AREA_RATIO] = \
            np.count_nonzero(edges_in_dilated_cc) / \
            np.count_nonzero(cc_dilated_mask)
  
        # flow phase bins to flow magnitude sums 'histogram' using dilated mask
        bin_sums, bin_sums_div_by_count, bin_sums_div_by_area = \
            get_flow_mag_in_phase_bins(flow, cc_dilated_mask,
                                       -np.pi / 8, np.pi / 4)

        
        features_vector[:,
                        FEATURE_FLOW_PHASE_BIN_MAG_SUM_HIST_DIV_COUNT_LEFT : \
                        FEATURE_FLOW_PHASE_BIN_MAG_SUM_HIST_DIV_COUNT_RIGHT \
                        + 1] = bin_sums_div_by_count

        # area
        area = np.count_nonzero(cc_dilated_mask)
        features_vector[:, FEATURE_AREA] = area
        
        # feature perimeter
        kernel = np.ones((3, 3))
        
        features_vector[:, FEATURE_PERIMETER] = \
            np.count_nonzero(cv2.subtract(cv2.dilate(cc_dilated_mask, kernel),
                                          cv2.erode(cc_dilated_mask, kernel)))
    
        # minimum area rectangle features
        
        # get the external contour points 
        contours = cv2.findContours(cc_dilated_mask, cv2.RETR_EXTERNAL,
                                    cv2.CHAIN_APPROX_SIMPLE)
        
        # reshape the points into N x 2 for passing to cv2.minAreaRect()
        contour_points = contours[1][0].reshape((contours[1][0].shape[0], -1))
        
        # get the minimum area rectangle info
        rect_center_point, rect_shape, rotation = \
            cv2.minAreaRect(contour_points)
        
        # separate the rotations
        width_along_rotated_x, height_along_rotated_y = rect_shape
        
        '''when width_along_rotated_x is smaller than height_along_rotated_y, 
        adjust the rotation to distinguish orientation'''
        if width_along_rotated_x < height_along_rotated_y:
            
            if abs(rotation) != 90:
                
                rotation = -rotation
                
            else:
                
                rotation = 0
                
        #-force -90 to become 90'''            
        elif rotation == -90:
            
            rotation = 90
    
        # save the orientation feature
        features_vector[:, FEATURE_MIN_AREA_RECT_ORIENT] = rotation
    
        # save the ratio of component area to minimum area rectangle
        features_vector[:, FEATURE_AREA_TO_MIN_AREA_RECT] = \
            area / (width_along_rotated_x * height_along_rotated_y)
            
        # save ratio of short side to long side of minimum area rectangle
        features_vector[:, FEATURE_MIN_AREA_RECT_SHORT_TO_LONG_SIDES_RATIO] = \
            min(rect_shape) / max(rect_shape)
            
        if min(rect_shape) < 0 or max(rect_shape) < 0:
            print(rect_shape)
            
        # save the features vector for the label
        cc[-1].append(features_vector)
        
    return cc

def classify(cc, model):
    '''
    cc - Output of cv2.connectedComponentsWithStats() with appended
         element holding feature vector for the components          
    model - trained model used for classification
    
    Return cc with another element holding the prediction (0 - toss) and
    (1 - keep) on the features for each component. The background component
    is forced to have a -1.
    
    '''
    prediction = model.predict(np.vstack(cc[-1][1:]))
    prediction = np.hstack([np.array([-1]), prediction])
    
    cc.append(prediction)
    
    return cc
    



def cc_to_mask(cc, cc_labels):
    '''
    This function a makes a mask from a connected component label matrix
    and a list of connected component labels of interest.
    
    cc - List holding output of cv2.connectedComponentsWithStats with 
         any appended information
    cc_labels - This list holds the connected component labels
    
    '''
    # make mask for CC(s) of interest for coloring
    mask_layer = np.any([cc[1] == v for v in cc_labels], axis=0)
    print('cc_to_mask', mask_layer.shape, mask_layer.dtype)
    mask = np.stack([mask_layer, mask_layer, mask_layer], axis=2)
    
    return mask


def overlay(frame, cc, cc_labels, color_bgr, alpha=0.25):
    '''
    This function gets uses a connected component label matrix and a list of
    connected component labels to overlay the connected components specifed
    with in the frame passed. The connected components are tinted the color
    passed in color_bgr.
    
    frame - holds the image array
    cc - List holding output of cv2.connectedComponentsWithStats with 
         any appended information
    cc_labels - This list holds the connected component labels of the
                CCs to be colored.
    color_bgr - color (blue, green, red color)
    alpha - weight of overlayed frame (default 0.25); the frame overlayed into
            will have weight 1 - alpha
    
    https://www.pyimagesearch.com/2016/03/07/transparent-overlays-with-opencv/
    '''
    
    # make a frame with three channels if frame has only one channel
    if len(frame.shape) == 2:
        
        frame = np.stack([frame, frame, frame], axis=2)
    
    # get the connected component mask for the connected components of interest
    mask = cc_to_mask(cc, cc_labels)

    # make the frame to overlay
    
    #initialize the overlay frame to be the same as frame
    overlay = np.copy(frame)
    
    # make a frame the color passed
    size = overlay.shape
    color_frame = np.stack([color_bgr[0] * np.ones(size[0:2], dtype=np.uint8),
                            color_bgr[1] * np.ones(size[0:2], dtype=np.uint8),
                            color_bgr[2] * np.ones(size[0:2], dtype=np.uint8)],
                           axis=2)
    
    # color the connected components of interest
    np.copyto(overlay, color_frame, where=mask)
            
    # mix
    result = cv2.addWeighted(frame, 1 - alpha, overlay, alpha, gamma=0)
    
    return (result, mask)