# -*- coding: utf-8 -*-
"""
Created on Fri Feb  9 23:14:59 2018

@author: Callie Darsey

I want a way to label components with a list of descriptors (for example,
hand, wrist, forearm, face, ear, other)

labeling_image_components_01
"""


import cv2
import numpy as np
import tkinter
from functools import partial
import tkinter_checkbox_example as menu

class LabelDescriptorMenu:
    '''
    Makes a menu of buttons or text entry and button for label description
    choice or entry. Activities entered via text entry are added as buttons.
    '''
    
    def __init__(self, activities_of_interest = ["talking",  "typing",
                "writing", "pointing", "hand gesture for concept description"],
                activities_not_of_interest = []):
        '''
        Initializer method.
        
        activities_of_interest - holds initial list of string descriptors of
                                 the activities that are of interest.
                                 
        activities_not_of_interest - holds initial list of string descriptors
                                    of the activities that are not of interest.                     
        '''
        # make the descriptor instance variable
        self.descriptor = ''
        
        # of interest descriptors
        self.of_interest = activities_of_interest
        
        # not of interest descriptors
        self.not_of_interest = []

        
    def _quit(self, descriptor='', get=False):
        '''
        This function saves the string in descriptor instance
        variable destroys the TK object.
        '''
        
        # get the descriptor entered
        if get:
            self.descriptor = self.entry.get()
                        
            # when the descriptor is not an empty string
            if len(self.descriptor):
                
                # save new descriptor
                if self.label_type == 1:
                    
                    if self.descriptor not in self.of_interest:
                        
                        self.of_interest.append(self.descriptor)
                    
                elif self.label_type == 0:
                    
                    if self.descriptor not in self.not_of_interest :
                        
                        self.not_of_interest.append(self.descriptor)
        else:
            self.descriptor = descriptor
            
        self.tk.destroy()        
    
    def _make_button(self, descriptor):
        '''
        make button
        '''
        func = partial(self._quit, descriptor=descriptor)
        button = tkinter.Button(self.frame, text=descriptor, command=func) 
        button.pack()       
    
    def display_menu(self, label_type):
        '''
        Display the menu and handle its demise.

        label_type - 0 or 1 (0 for 'not of interest' and 1 for 'of interest')
        ''' 
        
        # save type as instant variable
        self.label_type = label_type
        
        # make object
        self.tk = tkinter.Tk()
               
        # put a frame in it        
        self.frame = tkinter.Frame(self.tk, relief=tkinter.RIDGE, borderwidth=10) 
        self.frame.pack(fill=tkinter.BOTH, expand=1)
               
        
        if self.label_type == 1:
            
            # give info
            label = tkinter.Label(self.frame, text='of interest') 
            label.pack(fill=tkinter.X, expand=1)
            
            # Add a button for each activity of interest            
            for descriptor in self.of_interest:
                self._make_button(descriptor)
                
        elif self.label_type == 0:
            
            # give info
            label = tkinter.Label(self.frame, text='not of interest') 
            label.pack(fill=tkinter.X, expand=1)
            
            # Add button for activities not of interest if descriptor(s) present
            for descriptor in self.not_of_interest:
                self._make_button(descriptor)
        
        self.entry = tkinter.Entry(self.frame, bd = 5)
        self.entry.pack(side=tkinter.LEFT)        
        
        func = partial(self._quit, get=True)        
        button = tkinter.Button(self.frame, text="save", command=func) 
        button.pack(side=tkinter.RIGHT)
        
        self.tk.mainloop()

    def get_descriptor(self):
        '''
        Returns the current descriptor.
        '''
        return self.descriptor




class YesNoQuit():
    def __init__(self, question="...?", default='no', other_buttons=[]):
        '''
        Initialize the YesNo instance with a question and default answer.
        
        question - string with the question to ask
        default - string that is the default if the window is exited 
                  ('no' is default)
        other_buttons - list two element tuples holding the text to put on
                        the button and the string to answer that button
        '''
        self.question = question
        self.answer = default
        self.other_buttons = other_buttons
        
    def ask(self):
        '''
        This method causes the window with the yes no quit question to pop up.
        '''
        
        self.tk = tkinter.Tk()
        
        self.frame = tkinter.Frame(self.tk, relief=tkinter.RIDGE, borderwidth=10) 
        self.frame.pack(fill=tkinter.BOTH, expand=1)
        
        label = tkinter.Label(self.frame, text=self.question, justify=tkinter.LEFT) 
        label.pack(fill=tkinter.X, expand=1)
        
        for button_text, answer_text in self.other_buttons:
            self._make_button(button_text, answer_text)

        button1 = tkinter.Button(self.frame, text="yes", command=self._yes) 
        button1.pack(side=tkinter.TOP)

        button2 = tkinter.Button(self.frame,text="no", command=self._no) 
        button2.pack(side=tkinter.TOP)
        
        button3 = tkinter.Button(self.frame,text="quit", command=self._quit) 
        button3.pack(side=tkinter.TOP)
        
        self.tk.mainloop()
    
    def get_reply(self):
        '''
        Return the answer string.
        '''
        return self.answer
    
    def _make_button(self, button_text, answer_text):
        '''
        make button
        
        button_text = string with the text for the button
        answer_text = string to give as an anwer
        '''
        func = partial(self._other_button, answer_text)
        button = tkinter.Button(self.frame, text=button_text, command=func) 
        button.pack(side=tkinter.TOP)
        
    def _other_button(self, answer_text):
        '''
        callback method for other buttons
        
        answer_text - the string to reply with
        '''
        self.answer = answer_text
        self.tk.destroy()
            
    def _yes(self):
        '''
        callback method for a yes reply
        '''
        self.answer = 'yes'
        self.tk.destroy()
        
    def _no(self):
        '''
        callback method for a no reply
        '''
        self.answer = 'no'
        self.tk.destroy()
        
    def _quit(self):
        '''
        callback method for a quit reply
        '''
        self.answer = 'quit'
        self.tk.destroy()
        



class CCLabelingModified:
    ''' references
    https://www.pyimagesearch.com/2015/03/09/capturing-mouse-click-events-with-python-and-opencv/
    https://docs.opencv.org/3.3.1/d7/dfc/group__highgui.html
    '''
    def __init__(self, checkbox_strings):
        '''
        Initializer method
        
        checkbox_strings - list of descriptors (spaces sucht that they are 
                           they are the same length for the menu are added
                           within the __init__ method)
        '''
        
        # save instance variable for later reference
        self.checkbox_strings = checkbox_strings
        
        # make an instance of for the descriptor menu
        self.descriptor_menu = menu.ChooseDescriptorsSimple(checkbox_strings)
        
        # make an instance for a yes no quit menu
        other_button = [("make remaining ['', '', 'other']",
                         "make remaining ['', '', 'other']")]
        statement = "Not all the connected components (CCs) were labeled. \n" + \
                    "Do you really want to leave the frame for good?\n" + \
                    "\n" + \
                    "\tChoose 'make remaining ['', '', 'other'] to \n" + \
                    "\t\tlabel the remaining CCs ['', '', 'other'] and \n" + \
                    "\t\tleave frame and continue \n" + \
                    "\n" + \
                    "\tChoose 'yes' to leave frame and continue.\n" + \
                    "\n" + \
                    "\tChoose 'no' to stay with frame.\n" + \
                    "\n" + \
                    "\tChoose 'quit' to leave frame and quit.\n" + \
                    "\n" + \
                    "(continuing and quiting must be handled by the calling script)\n"
        self.yes_no_quit = YesNoQuit(question=statement, other_buttons=other_button)
        
        # make the down and up variables for later holding x, y coordinates
        self.down = []
        self.up = []
        
        
    def set_anew_run(self, frames, index, cc):
        '''
        This function initializes with the new parameters passed and run
        the classification labeling for the frames[index] frame.
        
        frames - list of frames
        index  - index into frames indicating the frame over which we are
                 labeling connected components
        cc     - list holding output of cv2.connectedComponentsWithStats with 
                 any appended information
        '''
        # save what is passed in
        self.frames = frames
        self.index = index
        self.cc = cc

        # initialize space in cc for info collected when labeling
        self._prep_label_space()
        
        # initialize the current index length of frames
        self.cur_index = index
        self.flen = len(frames)

        # display the frame
        self.title = 'Press escape to do classification labeling on next frame'
        
        # set the window
        cv2.namedWindow(self.title, cv2.WINDOW_GUI_NORMAL | cv2.WINDOW_AUTOSIZE)
        
        # save the frame of interest being uncolored
        self.frame_uncolored = self.frames[self.index].copy()
                               
        #  for the first time, color the CC's blue
        if np.unique(self.cc[1])[1:].shape != (0,):
            self._color_cc(np.unique(self.cc[1])[1:], [255, 0, 0], 0.5)
        
        while True:
            
            cv2.imshow(self.title, self.frames[self.index])
            
            cv2.moveWindow(self.title, 0, -40)
            
            # set the mouse callback
            cv2.setMouseCallback(self.title, self._click)
                
            k = cv2.waitKey(0) & 0xff
                
            if k == 27:
                
                # verify that your really want to leave
                if not self._all_cc_labeled():
                    
                    # verify with user
                    self.yes_no_quit.ask()
                    self.reply = self.yes_no_quit.get_reply()
                    
                    if self.reply == 'yes' or self.reply == 'quit':
                        
                        # close all the OpenCV windows
                        cv2.destroyAllWindows()
                        
                        break
                    
                    elif self.reply == "make remaining ['', '', 'other']":

                        # make all the targets ['', '', 'other']
                        empty_indices = np.where(self.empty)[0]
                        
                        for empty_index in empty_indices:
                            cc[-1][empty_index] = ['', '', 'other']

                        # color them red
                        if empty_indices.shape != (0,):
                            self._color_cc(empty_indices, (0, 0, 255), 0.5)
                        
                        # close all the OpenCV windows
                        cv2.destroyAllWindows()
                        
                        break                        
                else:
                    
                    self.reply = 'yes'
        
                    # close all the OpenCV windows
                    cv2.destroyAllWindows()
                    
                    break
        
        
                
    def get_results(self):
        '''
        This method returns the updated frames list and updated cc structure 
        (without the clicks) and either the yes or quit flag which can be used
        to either continue or quit.
        '''
        
        return (self.frames, self.cc, self.reply)        
        

        
    def _prep_label_space(self):
        '''
        This function adds two more entries to the self.cc list. The first will 
        hold the numberic labels 1 - 'of interest' and 0 - 'not of interest'. 
        The second will hold a more descriptive string (or just be an empty
        string).
        '''
        
        # descriptive string (descriptor)
        self.cc.append([])
        self.cc.append([[] for i in range(self.cc[0])])
        
        '''
        cc[-1] - (number of labels) list of lists each holding the target
                 (an example of a target is ['hand or forearm', 'face', ''])
        '''
        
        
#        # classification label value container
#        self.cc.append(np.full_like(self.cc[3][:,0], np.nan, dtype=np.float))
#        
#        # background 'not of interest'
#        self.cc[-1][0] = 0
#        
#        # descriptive string (descriptor)
#        self.cc.append(['' for i in range(self.cc[0])])
#        
#        # The 0 classification label is typically the background
#        self.cc[-1][0] = 'background'
#        
#        # number of clicks on that component
#        self.cc.append(np.full_like(self.cc[3][:,0], 0, dtype=np.int))
#        
#        '''
#        cc[-3] - (number of labels) x 1 numpy array of for classification labels
#        cc[-2] - (number of labels) list of descriptor strings
#        cc[-1] - (number of labels) x 1 numpy array of for number of clicks per CC
#        '''
        
    def _color_cc(self, cc_labels, color_bgr, alpha=0.25):
        '''
        This function colors CCs in the image.
        
        frame - holds the image array
        cc_labesl - This list holds the connected component labels of the
                    CCs to be colored.
        color_bgr - color (blue, green, red color)
        
        https://www.pyimagesearch.com/2016/03/07/transparent-overlays-with-opencv/
        '''

        # make a frame colored; start by initializing with current frame
        if len(self.frame_uncolored.shape) == 2:
            
            self.frame_uncolored = np.stack([self.frame_uncolored,
                                             self.frame_uncolored,
                                             self.frame_uncolored],
                                            axis=2)
        if len(self.frames[self.index].shape) == 2:

            self.frames[self.index] = np.stack([self.frames[self.index],
                                                self.frames[self.index],
                                                self.frames[self.index]],
                                               axis=2).copy()

        # make mask for CC(s) of interest for coloring
        mask_layer = np.any([self.cc[1] == v for v in cc_labels], axis=0)
        mask = np.stack([mask_layer, mask_layer, mask_layer], axis=2)
        
        # uncolor CC(s) of interest
        np.copyto(self.frames[self.index], self.frame_uncolored, where=mask)
    
        # make another frame with the CCs fully colored for overlaying
        overlay = self.frames[self.index].copy()
        size = overlay.shape
        color_frame = np.stack([color_bgr[0] * np.ones(size[0:2], dtype=np.uint8),
                                color_bgr[1] * np.ones(size[0:2], dtype=np.uint8),
                                color_bgr[2] * np.ones(size[0:2], dtype=np.uint8)],
                               axis=2)
        np.copyto(overlay, color_frame, where=mask)
                
        # mix
        self.frames[self.index] = cv2.addWeighted(self.frames[self.index],
                                                  1 - alpha, overlay, alpha, 0)
    
    
    def _verify_click(self):
        '''
        This function verifies the click was within one connected component.
        '''
        
        # make rectangle mask defined by the down and up click locations
        click_rect = np.zeros(self.cc[1].shape, dtype=np.int32)
        color = 1 # grayscale
        thickness = -1 # -1 for filled
        cv2.rectangle(click_rect, self.down, self.up, color, thickness)
        
        # area of click rectangle mask
        area_click_rect = np.count_nonzero(click_rect)
        
        # area of overlap of CC labels and click rectangle
        overlap = self.cc[1] * click_rect
        area_overlap = np.count_nonzero(overlap)

        # report the results        
        if area_click_rect != area_overlap:
            
            # warn on the console
            print('bad click, try again')
            
            # report an invalid click
            return False
        else:
            # extract the clicked CC label in label matrix
            self.cc_label = np.unique(overlap)[1]
            
            # report a valid click
            return True
            
    def _all_cc_labeled(self):
        '''
        This function checks whether all the CCs got assigned a classification
        label.
        '''
        
        # check if any empty targets are left
        self.empty = [False]
        for target in self.cc[-1][1:]:
            
            if len(target) == 0:
                self.empty.append(True)
            else:
                self.empty.append(False)
            
        if np.any(self.empty):
            
            # not all the CCs got classification labels
            return False
        else:
            
            # all the CCs got classification labels
            return True
        
    def _process_CC_click(self, color_bgr):
        '''
        This function processes a component click.
        
#        label_type - 0 or 1 (0 for 'not of interest' and 1 for 'of interest')
        color_bgr - color (blue, green, red color)
        '''
        
#        # check that the label_type is 0 or 1
#        assert label_type == 1 or label_type == 0
        
        # verify click is within CC
        if self._verify_click() == True:
            
            # prepare the checkbox status list
            if len(self.cc[-1][self.cc_label]) == 0:
                check_on = [False] * len(self.checkbox_strings)
            else:
                check_on = []
                for string in self.cc[-1][self.cc_label]:
                    if string == '':
                        check_on.append(False)
                    else:
                        check_on.append(True)

            self.descriptor_menu.make_menu(check_on)
            
            target = self.descriptor_menu.get_target()
            
            self.cc[-1][self.cc_label] = target
            
            self._color_cc([self.cc_label], color_bgr, 0.5)
            
            # display
            cv2.imshow(self.title, self.frames[self.index])
            
#            # either save the info or display the descriptor menu
#            self.cc[-1][self.cc_label] += 1
#            
#            # on even clicks, just update the classification label
#            if np.mod(self.cc[-1][self.cc_label], 2) == 1:
#                
#                # update the classification label
#                self.cc[-3][self.cc_label] = label_type
#                
#                # show the image with the component colored
#                self._color_cc([self.cc_label], color_bgr)
#                
#                # display
#                cv2.imshow(self.title, self.frames[self.index])
#            
#            # on odd clicks, display the descriptor menu
#            else:
#                
#                # open the descriptor menu
#                self.descriptor_menu.display_menu(label_type)
#                
#                # get descriptor result
#                descriptor = self.descriptor_menu.get_descriptor()
#                
#                # update the descriptor result
#                self.cc[-2][self.cc_label] = descriptor        


        
    def _click(self, event, x, y, flags, params):
        
        if event == cv2.EVENT_LBUTTONDOWN:
            #print('left mouse down')
            
            # save coordinates and wait to come up to process
            self.down = (x, y)
            
        elif event == cv2.EVENT_LBUTTONUP:
            #print('left mouse up')
            
            # save label (1) and imshow and such if click is valid 
            self.up = (x, y)
            
            # verify at frame in which we are making classification labels
            if self.cur_index == self.index:
                
                # process the click
                self._process_CC_click([0, 255, 0])

        elif event == cv2.EVENT_RBUTTONDOWN:
            #print('right mouse down')
            
            # save coordiantes and wait to come up to process
            self.down = (x, y)
            
        elif event == cv2.EVENT_RBUTTONUP:
            #print('right mouse up')
            
            # save label (1) and imshow and such if click is valid 
            self.up = (x, y)
            
            # verify at frame in which we are making classification labels
            if self.cur_index == self.index:
                
                # process the click
                self._process_CC_click([0, 0, 255])            
                        
        elif event == cv2.EVENT_MBUTTONUP:
            #print('middle up')
            
            # display next (unprocessed) frame if there is one
            if self.cur_index + 1 < self.flen:
                
                # update the index
                self.cur_index += 1
                    
                cv2.imshow(self.title, self.frames[self.cur_index])
                                        
        elif event == cv2.EVENT_MOUSEWHEEL:
            #print('scrolled mouse wheel')
            
            # when the scroll was down (negtive flag value)
            if flags < 0:
                
                # display previous frame if there is one
                if self.cur_index - 1 >= 0:
                    
                    # update the index
                    self.cur_index += -1
                    
                    # display the frame
                    cv2.imshow(self.title, self.frames[self.cur_index])

       