# -*- coding: utf-8 -*-
"""
Created on Fri Feb  2 16:14:47 2018

@author: Callie Darsey (cj.darsey@gmail.com)

I want a function to overlay an mask in another image.

I want a function that overlays a smaller mask in the upper right corner of
a larger image.

"""

import numpy as np
import cv2

def cc_to_mask(cc, cc_labels):
    '''
    This function a makes a mask from a connected component label matrix
    and a list of connected component labels of interest.
    
    cc - List holding output of cv2.connectedComponentsWithStats with 
         any appended information
    cc_labels - This list holds the connected component labels
    
    '''
    # make mask for CC(s) of interest for coloring
    mask_layer = np.any([cc[1] == v for v in cc_labels], axis=0)
    mask = np.stack([mask_layer, mask_layer, mask_layer], axis=2)
    
    return mask


def overlay(frame, cc, cc_labels, color_bgr, alpha=0.25):
    '''
    This function gets uses a connected component label matrix and a list of
    connected component labels to overlay the connected components specifed
    with in the frame passed. The connected components are tinted the color
    passed in color_bgr.
    
    frame - holds the image array
    cc - List holding output of cv2.connectedComponentsWithStats with 
         any appended information
    cc_labels - This list holds the connected component labels of the
                CCs to be colored.
    color_bgr - color (blue, green, red color)
    alpha - weight of overlayed frame (default 0.25); the frame overlayed into
            will have weight 1 - alpha
    
    https://www.pyimagesearch.com/2016/03/07/transparent-overlays-with-opencv/
    '''
    
    # make a frame with three channels if frame has only one channel
    if len(frame.shape) == 2:
        
        frame = np.stack([frame, frame, frame], axis=2)
    
    # get the connected component mask for the connected components of interest
    mask = cc_to_mask(cc, cc_labels)

    # make the frame to overlay
    
    #initialize the overlay frame to be the same as frame
    overlay = np.copy(frame)
    
    # make a frame the color passed
    size = overlay.shape
    color_frame = np.stack([color_bgr[0] * np.ones(size[0:2], dtype=np.uint8),
                            color_bgr[1] * np.ones(size[0:2], dtype=np.uint8),
                            color_bgr[2] * np.ones(size[0:2], dtype=np.uint8)],
                           axis=2)
    
    # color the connected components of interest
    np.copyto(overlay, color_frame, where=mask)
            
    # mix
    result = cv2.addWeighted(frame, 1 - alpha, overlay, alpha, gamma=0)
    
    return result

def form_number(number, digits, size):
    '''
    This function returns a one channel image with the number in the upper
    right corner.
    
    number - integer number
    digits - dictionary with keys as strings of the digits 0, 1, 2, 3, 4, 5,
             6, 7, 8, 9 and the values as the array of the digit extracted from
             a label matrix which had all digits represented.
    size - two element list or tuple with the dimensions (rows, columns) to
           to give the output image
    '''
    
    assert len(size) == 2, \
        'In form_number, size argument was expected to have two elements.'
    
    # turn the number into a string
    number_str = str(number)
    
    # fill the label matrix with zeros
    label_matrix = np.zeros(size, dtype=np.uint8)
    
    # use col_ind to keep track of the free space column-wise; 4 pixels spacing
    space = 4
    col_ind = size[1] - 1 - space
    
    # place digits
    for i in -(1+np.arange(len(str(number_str)))):
        
        # place a digit
        number = number_str[i]
        digit = digits[number]
        
        label_matrix[space : digit.shape[0] + space, 
                     col_ind - digit.shape[1] : col_ind] = digit
        
        # update column index
        col_ind = col_ind - digit.shape[1] - space
        
        # check
        if col_ind < 0:
            
            print('form number: number to big to fit')
            break
    return label_matrix   
        
def get_digits(digits_image='numbers.png'):
    '''
    This function returns the dictionary with the keys as strings of the
    digits 0 through 9 and the values the extracted digit from a label
    matrix gotten from the digits_image passed. The processing was based for
    the particular image with the digits I found online.
    
    digits_image - path to a image with the digits
    '''        
    
    # read the image
    numbers = cv2.imread(digits_image)
    
    # change to grayscale
    numbers_gray = cv2.cvtColor(numbers, cv2.COLOR_BGR2GRAY)
    
    # resize it to be smaller
    scale_factor = 0.25
    numbers_resize = cv2.resize(numbers_gray, (0,0),
                                fx=scale_factor, fy=scale_factor)
    
    # threshold
    threshold = 200
    maxval = 255
    thresh, numbers_thresholded = cv2.threshold(numbers_resize, threshold, maxval,
                                                cv2.THRESH_BINARY_INV)
    
    # clean and shape with morphology operations
    numbers_cleaned = cv2.morphologyEx(numbers_thresholded, cv2.MORPH_OPEN,
                                       kernel=np.ones(3), iterations=1)
    numbers_cleaned = cv2.erode(numbers_cleaned, kernel=np.ones((1,3)))
    
    # get the connected components (CCs) and CC info
    connectivity=4
    cc = cv2.connectedComponentsWithStats(numbers_cleaned, connectivity,
                                          cv2.CV_32S)
    
    # extract each digit label matrix and store in dictionary
    digits = {}
    label_matrix = cc[1]
    cc_info = cc[2]
    
    for d in range(10):
        
        label = d+1
        
        left = cc_info[label, cv2.CC_STAT_LEFT]
        top = cc_info[label, cv2.CC_STAT_TOP]
        width = cc_info[label, cv2.CC_STAT_WIDTH]
        height = cc_info[label, cv2.CC_STAT_HEIGHT]
        
        digits[str(d)] = label_matrix[top : top  + height, left : left + width]
    
    return digits

if __name__ == '__main__':
    
    # example using the functions defined in this script
    
    digits = get_digits()
    
    for digit in digits.values():
        
        print('digit.shape', digit.shape)
        
        cv2.imshow('numbers', np.square(digit + 4).astype(np.uint8))
        
        print(cv2.waitKey(1))
        
        cv2.destroyAllWindows()
    
    image = cv2.imread('./video_copies_clips_track/copies_and_clips/' + \
                       'STI WA Apr25 A/STI WA Apr25 AMVI_0068_trans/' + \
                       'images/STI WA Apr25 AMVI_0068_trans__860_910__00073.png')
    
    placed = form_number(9876543210, digits, image.shape[0:2])
    
    connectivity = 4
    cc_placed = cv2.connectedComponentsWithStats(placed, connectivity, cv2.CV_32S)
    
    
    marked = overlay(image, cc_placed, np.unique(cc_placed[1])[1:], (0, 255, 0),
                     alpha=1)
    
    cv2.imshow('marked', marked)
    
    cv2.waitKey(0)
    
    cv2.destroyAllWindows()