# -*- coding: utf-8 -*-
"""
Created on Thu Feb  1 09:07:10 2018

@author: Callie Darsey

I want a skin detector.

Taking this from Wenjing's Thesis.

"""

import cv2
import numpy as np

def DetectSkinRegions(BGR):
    '''
    This function returns a mask of detected skin regions.
    
    This function comes from "Human Attention Detection Using AM-FM
    Representations" by Wenjing Shi (for a University of New Mexico thesis),
    and Shi references the following:
        
    Nusirwan Anwar bin Abdul Rahman, Kit Chong Wei, and John See. Rgb-h-cbcr
    skin colour model for human face detection. Faculty of Information
    Technology, Multimedia University, 2007.
    
    From Shi's thesis, "This function is used to construct bounding
    rules for skin regions detected based on four color space models:
        RGB_1, RGB_2, HSV, and YCbCr."
    
    BGR - image array    
    '''
    
    # break into blue, green, red for ease of reading
    B = BGR[:,:,0]
    G = BGR[:,:,1]
    R = BGR[:,:,2]
    
    # turn type into double for ease of reading
    B_double = B.astype(np.float64)
    G_double = G.astype(np.float64)
    R_double = R.astype(np.float64)
          
    # get BGR_1 model
    first_AND_parts = [B > 20, G > 40, R > 60, R > G, R > B, 
                          10 < np.absolute(R - G), np.absolute(R - G) < 45, 
                          np.absolute(R - B) < np.absolute(R - G)]
    
    second_AND_parts = [np.absolute(R - G) < 45, np.absolute(R - B) > 10, 
                           np.absolute(R - G) < np.absolute(R - B)]
        
    BGR_1 = np.logical_or(np.all(first_AND_parts, axis=0),
                          np.all(second_AND_parts, axis=0))
    
    # get BGR_2 model
    '''
    adding small value two (R + G + B) divisor to avoid divide by zero warning
    '''
    eps = 1e-10
    AND_parts = [0.36 <= (R_double / (R_double + G_double + B_double +eps)),
                 (R_double / (R_double + G_double + B_double + eps)) <= 0.44,
                 0.2 <= (G_double / (R_double + G_double + B_double + eps)),
                 (G_double / (R_double + G_double + B_double + eps)) <= 0.36]
    
    BGR_2 = np.all(AND_parts, axis=0)
    
    
    # get HSV    
    HSV = cv2.cvtColor(BGR, cv2.COLOR_BGR2HSV)
    
    # break into hue, saturation, and value for ease of reading
    H = HSV[:,:,0]
    S = HSV[:,:,1]
    V = HSV[:,:,2]
    
    # get HSV model
    '''
    H, S, and V are scaled to be in [0, 255] np.uint8, see 
    https://docs.opencv.org/2.4.2/modules/imgproc/doc/miscellaneous_transformations.html#cvtcolor.
    thus using 0 <= H <= 255 (instead of 0 <= H <= 1), using 26 <= S <= 77
    (instead of 0.1 <= S <= 0.3), and using 51 <= V <= 204 (instead of
    0.2 <= V <= 0.8)    
    '''
    AND_parts = [0 <= H, H <= 255, 26 <= S, S <= 77, 51 <= V, V <= 204 ]
    
    HSV_1 = np.all(AND_parts, axis=0)
    
    # get YCrCb    
    YCrCb = cv2.cvtColor(BGR, cv2.COLOR_BGR2YCrCb)
    
    # break of the red-difference and blue-difference chroma components
    Cr = YCrCb[:,:,1]
    Cb = YCrCb[:,:,2]
    
    # get YCrCB model
    AND_parts = [135 <= Cr, Cr <= 145, 110.5 <= Cb, Cb <= 135.5]
    
    YCrCb_1 = np.all(AND_parts, axis=0)
    
    # return final model
    return np.all([BGR_1, BGR_2, HSV_1, YCrCb_1], axis=0).astype(np.uint8)*255

  
if __name__ == '__main__':
   
    BGR = cv2.imread('girl_pic_from_Wenjing_thesis.png')
    
    skin_mask = DetectSkinRegions(BGR)
    
    BGR_skin_mask = cv2.bitwise_and(BGR, BGR, mask=skin_mask)
    
    cv2.imshow('color image', BGR)
    
    cv2.imshow('masked', BGR_skin_mask)
    
    cv2.waitKey(0)
    
    cv2.destroyAllWindows()    