# -*- coding: utf-8 -*-
"""
Created on Sun Feb 11 13:27:39 2018

@author: Callie Darsey (cj.darsey@gmail.com)

I want to figure out what the feature from a connected component should be.

So here is a toy problem.

toy_problem_place_cc
"""

import numpy as np
import cv2
import imutils

def centroid_center_crop(frame, cc, cc_label):
    '''
    This function crops connected component(s) such that the centroid is in 
    the middle of the crop (zero padding is added to achieve this if
    necessary).
    
    frame - frame the mask will extract information from
    cc - list holding output of cv2.connectedComponentsWithStats with
         any appended information 
    cc_label - connected component label being cropped
    '''
    
    # get out the one component
    one_cc = np.zeros_like(cc[1])
    one_cc[cc[1] == cc_label] = cc_label
    
    # mask the portion of frame aligned with the connected component
    portion = cv2.bitwise_and(frame, frame, mask=(one_cc > 0).astype(np.uint8))
    
    # centroid x and y forced to int
    c_col = int(cc[3][cc_label, 0])
    c_row = int(cc[3][cc_label, 1])
    
    # points of the bounding box.
    pt1_col = cc[2][cc_label, cv2.CC_STAT_LEFT]
    pt1_row = cc[2][cc_label, cv2.CC_STAT_TOP]
    
    pt2_col = pt1_col + cc[2][cc_label, cv2.CC_STAT_WIDTH] - 1
    pt2_row = pt1_row + cc[2][cc_label, cv2.CC_STAT_HEIGHT] - 1
    
    # left and right column distance
    left_col_dist = c_col - pt1_col
    right_col_dist = pt2_col - c_col
    
    # left and right row distance
    upper_row_dist = c_row - pt1_row
    lower_row_dist = pt2_row - c_row
    
    # row and column with of one side of rectangle excluding center
    col_dist = max([left_col_dist, right_col_dist])
    row_dist = max([upper_row_dist, lower_row_dist])
    
    # extract the portion of the frame associated with connected component
    extracted = portion[pt1_row:pt1_row + cc[2][cc_label, cv2.CC_STAT_HEIGHT],
                        pt1_col:pt1_col + cc[2][cc_label, cv2.CC_STAT_WIDTH]]
    
    # place the croped component with its centroid in the center
    crop = np.zeros((2 * row_dist + 1, 2 * col_dist + 1), dtype=np.uint8)
    crop[row_dist - upper_row_dist : \
         row_dist - upper_row_dist + cc[2][cc_label, cv2.CC_STAT_HEIGHT], \
         col_dist - left_col_dist : \
         col_dist - left_col_dist + cc[2][cc_label, cv2.CC_STAT_WIDTH], ...] = \
         extracted
         
    return crop
         
    
    
    
def insert_centroid_centered_crop(centroid_centered_crop, new_dim):
    '''
    Insert image into a new image of dimension new_dim.
    
    centroid_centered_crop - crop image with the centroid at the center
    new_dim - 2 element or three element size for the new image (the
              dimensions). First two dimensions are expected to be odd.
    '''
    
    assert len(centroid_centered_crop.shape) == len(new_dim)
    assert new_dim[0] % 2 == 1 and new_dim[1] % 2 == 1
    assert centroid_centered_crop.shape[0] % 2 == 1
    assert centroid_centered_crop.shape[1] % 2 == 1
    
    old_height, old_width = centroid_centered_crop.shape[0:2]
    
    new_height, new_width = new_dim[0:2]
    
    new = np.zeros(new_dim, dtype=np.uint8)
    
    new[int(new_height/2) - int(old_height/2) : \
        int(new_height/2) - int(old_height/2) + old_height,
        int(new_width/2) - int(old_width/2) : \
        int(new_width/2) - int(old_width/2) + old_width, ...] = \
        centroid_centered_crop
        
    return new
    
    
if __name__ == '__main__':
    
    m = np.array([[   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0],
                  [   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0],
                  [   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0],
                  [   0,   0,   0, 255, 255,   0,   0,   0,   0,   0,   0],
                  [   0,   0,   0,   0, 255, 255,   0,   0,   0,   0,   0],
                  [   0,   0, 255, 255, 255, 255,   0, 255,   0,   0,   0],
                  [   0,   0,   0,   0, 255, 255, 255, 255, 255,   0,   0],
                  [   0,   0,   0, 255, 255, 255, 255, 255, 255,   0,   0],
                  [   0,   0,   0,   0, 255, 255,   0,   0, 255, 255,   0],
                  [   0,   0,   0,   0,   0,   0, 255,   0,   0,   0,   0],
                  [   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0]], dtype=np.uint8)
    scaleFactor=1
    m = cv2.resize(m, (0,0), fx=scaleFactor, fy=scaleFactor)
    f = np.random.randint(0, 190, m.shape, dtype=np.uint8) + 66
    
    cc = cv2.connectedComponentsWithStats(m)
    
    m_draw = cv2.rectangle(np.stack([m, m, m], axis=2),
                           (cc[2][1, cv2.CC_STAT_LEFT], cc[2][1, cv2.CC_STAT_TOP]),
                           (cc[2][1, cv2.CC_STAT_LEFT] + cc[2][1, cv2.CC_STAT_WIDTH] - 1,
                            cc[2][1, cv2.CC_STAT_TOP] + cc[2][1, cv2.CC_STAT_HEIGHT] - 1),
                            (0, 255, 0), 1)
                          
    pt = (int(cc[3][1, 0]), int(cc[3][1, 1]))     
    print((cc[3][1, 0], cc[3][1, 1]), '-->',pt)                  
    cv2.line(m_draw, pt, pt, (0, 0, 255), 1)      
    
          
    centered_masked_f = centroid_center_crop(f, cc, np.unique(cc[1])[1])
    
    new_dim = (11, 11)
    
    sized = insert_centroid_centered_crop(centered_masked_f, new_dim)
    
    cv2.imshow('title1', m)
    cv2.imshow('title2', m_draw)
    cv2.imshow('title3', f)
    cv2.imshow('title4', centered_masked_f)
    cv2.imshow('title5', sized)
    cv2.waitKey(0)
    cv2.destroyAllWindows()



